// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// auxx.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// some global macros                                                          |
// some global constants                                                       |
// some global typedefs                                                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_auxx_h
#define included_auxx_h

#ifdef C_AND_F
extern "C" {
#ifndef included_stdlib_h
#include <stdlib.h>
#define included_stdlib_h
#endif

#ifndef included_stdio_h
#include <stdio.h>
#define included_stdio_h
#endif
}
#else

#ifndef included_iostream
#include <iostream>
#define included_iostream
#endif

#endif

////////////////////////////////////////////////////////////////////////////////
// 1. INTERNAL PRECISION
//
//
// 1.1 from math.h:
//

extern "C" {
  double sqrt (double);
  double exp  (double);
  double log  (double);
  float  sqrtf(float);
  float  expf (float);
  float  logf (float);
  double pow  (double,double);
//   float  powf (float,float);
}

#ifdef DOUBLEPRECISION
//
// 1.2 double precision: 64bit floating point numbers to hold a real
//TM
namespace fasttree {
  typedef double real;
}

#define SQRT sqrt
#define POW  pow
#define EXP  exp
#define LOG  log

#else // DOUBLEPRECISION
//
// 1.3 single precision: 32bit floating point numbers to hold a real
//
namespace fasttree {
  typedef float real;
}

#ifdef EBUG
#define SQRT(x)   real(sqrt(x))
#define EXP(x)    real(exp(x))
#define LOG(x)    real(log(x))
#else
#define SQRT      sqrtf
#define EXP       expf            // use on linux machines
#define LOG       logf            // use on linux machines
// #define EXP(x)    real(exp(x))       // use on suns
// #define LOG(x)    real(log(x))       // use on suns
#endif
#define POW(x,y)  real(pow(double(x),double(y)))

#endif // DOUBLEPRECISION

////////////////////////////////////////////////////////////////////////////////
//
// 2. DATA TRANSFER VIA BODIES OR ARRAYS & PRECISION USED IN ARRAY ARGS

#ifdef C_AND_F
#undef USE_BODIES
#endif // C_AND_F

#ifdef USE_BODIES

#else  // USE_BODIES

namespace fasttree {
#ifdef DOUBLEINPUT
  typedef double areal;
#else  // DOUBLEINPUT
  typedef float areal;
#endif // DOUBLEINPUT
}

#endif // USE_BODIES

////////////////////////////////////////////////////////////////////////////////
//
// 3. DIMENSIONALITY
//
// 3.1 Two dimensions (disk)
//
#ifdef TWODIMENSIONAL

#define NDIM 2   // # dimensions
#define NSUB 4   // 2^NDIM

#ifndef included_paiR_h
#include <tree/paiR.h>
#endif

namespace fasttree {
  typedef class paiR<real>       vect;
  typedef real                   amom;
}

#else // TWODIMENSIONAL
//
// 3.2 Three dimensions
//

#define NDIM 3   // # dimensions
#define NSUB 8   // 2^NDIM

#ifndef included_trip_h
#include <tree/trip.h>
#endif

namespace fasttree {
  typedef class tripel<real>     vect;
  typedef vect                   amom;
}

#endif // TWODIMENSIONAL

////////////////////////////////////////////////////////////////////////////////
//
// 4. USEFUL CONSTANTS AND TYPEDEFS
//

namespace fasttree {
  typedef unsigned short   indx;
  typedef unsigned int     uint;
#ifdef C_AND_F
  inline void error  (const char*m) {printf(" TREE ERROR: %s",m);   exit(1); }
  inline void warning(const char*m) {printf(" TREE WARNING: %s",m); exit(1); }
#else
  inline void error  (const char*m) {std::cerr<<" TREE ERROR: "  <<m<<std::endl;
                                     exit(1); }
  inline void warning(const char*m) {std::cerr<<" TREE WARNING: "<<m<<std::endl;}
#endif
  const   real zero    = 0., 
               third   = 0.3333333333333333333,
               half    = 0.5,
               one     = 1.,
               two     = 2.,
               ten     = 10.;
}

////////////////////////////////////////////////////////////////////////////////

#endif  // included_auxx_h
