// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// buil.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class builder              see file buil.cc                                 |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_buil_h
#define included_buil_h

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  struct box;                                       // forward declare box      
  struct leaf;                                      // forward declare leaf     
  class  body;                                      // forward declare body     
  class  soul;                                      // forward declare soul     
  class  cell;                                      // forward declare cell     
  template<class BOX> class manager;                // forward declare manager  
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::builder                                                    
  //////////////////////////////////////////////////////////////////////////////
  class builder {
  private:
    uint           NLEAFS, NCRIT, NZOMBIES;
    int            DMAX;
    leaf          *L0, *LN;
    vect           XMIN, XMAX;
    manager<box>  *BM;
#ifndef USE_BODIES
    const areal *MASSES;
#endif
    box* new_box    (const uint);
    box* make_subbox(const box*, const indx, const uint);
    box* make_subbox(const box*, const indx, leaf*, const uint);
    void split_box  (box*, const indx, const uint);
    void addleafs   (box*);
    indx link_tree  (box*, cell*, cell*&, soul*&);
    bool is_twig    (const box*) const;
    void de_zombie  (box* &);
    //  void de_zombie   (cell* &);
  public:
#ifdef USE_BODIES
    builder(const uint,                             //  I:  N_crit              
	          body*,                            //  I:  array with bodies   
	    const uint);                            //  I:  # bodies            
#else
    builder(const uint,                             //  I:  N_crit              
	    const areal* [NDIM],                    //  I:  array with positions
	    const areal*,                           //  I:  array with masses   
	    const int*,                             //  I:  array with flags    
	    const uint);                            //  I:  size of arrays      
#endif
    ~builder();
    const uint& N_zombies  () const { return NZOMBIES; }
    const uint& N_allocated() const;                // # boxes allocated        
    const uint& N_used     () const;                // # boxes used             
    uint        N_needed   () const { return N_used() - N_zombies(); }
    uint        N_free     () const { return N_allocated()-N_used(); }
    const uint& N_leafs    () const { return NLEAFS; }
    const uint& Ncrit      () const { return NCRIT; }
    void build_tree        (const vect* = 0,        // [I]: center for root     
			    const int   = 50);      // [I]: max allowed depth   
    void link_tree         (cell*, soul*);          // link cells & souls       
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_buil_h
