// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// et2D.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class TY> class epaiR                  fully inline                |
// template<class TY> class esym22                 fully inline                |
// template<class TY> class esym222                fully inline                |
// some global functions                           fully inline                |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_et2D_h
#define included_et2D_h

#ifndef included_tn2D_h
#include <tree/tn2D.h>
#endif

////////////////////////////////////////////////////////////////////////////////

#define CTY         const TY
#define CI          const int
#define CUI         const unsigned int
#define CUS         const unsigned short
#define PAI         paiR<TY>
#define CPA         const paiR<TY>
#define SYM         sym22<TY>
#define CSY         const sym22<TY>
#define SYT         sym222<TY>
#define CST         const sym222<TY>
#define OP          operator
#define TM          template<class TY> inline

namespace fasttree {

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::epaiR for a paiR using external memory                     
  //////////////////////////////////////////////////////////////////////////////

#define EPA         epaiR
#define CEP         const epaiR

#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1];
#define FTX(op)     a[0] op x;    a[1] op x;


  template<class TY>
  class EPA {
  private:
    TY *a;
    EPA         (CEP &);                                      // disabled!
  public:
    CTY& OP[]   (CI  i)    const { return a[i]; }
    CTY& OP[]   (CUI i)    const { return a[i]; }
    CTY& OP[]   (CUS i)    const { return a[i]; }
    TY&  OP[]   (CI  i)          { return a[i]; }
    TY&  OP[]   (CUI i)          { return a[i]; }
    TY&  OP[]   (CUS i)          { return a[i]; }
    EPA         ( TY*b)          { a=b; }
         OP TY* ()               { return a; }
         OP CTY*()         const { return a; }
    EPA& OP=    (CTY x)          { FTX(=)     return *this; }
    EPA& OP*=   (CTY x)          { FTX(*=)    return *this; }
    EPA& OP/=   (CTY x)          { FTX(/=)    return *this; }
    PAI  OP*    (CTY x)    const { PAI y((CTY*)(*this)); y*=x; return y; }
    PAI  OP/    (CTY x)    const { PAI y((CTY*)(*this)); y/=x; return y; }
    EPA& OP=    (CPA&c)          { FT(a,c,=)  return *this; }
    EPA& OP+=   (CPA&c)          { FT(a,c,+=) return *this; }
    EPA& OP-=   (CPA&c)          { FT(a,c,-=) return *this; }
    PAI  OP+    (CPA&c)    const { PAI y(a); y+=c; return y; }
    PAI  OP-    (CPA&c)    const { PAI y(a); y-=c; return y; }
    TY   OP^    (CPA&c)          { return a[0]*c[1] - a[1]*c[0]; }
    EPA& OP=    (CEP&c)          { FT(a,c.a,=)  return *this; }
    EPA& OP+=   (CEP&c)          { FT(a,c.a,+=) return *this; }
    EPA& OP-=   (CEP&c)          { FT(a,c.a,-=) return *this; }
    PAI  OP+    (CEP&c)    const { PAI y(*a); y+=c.a; return y; }
    PAI  OP-    (CEP&c)    const { PAI y(*a); y-=c.a; return y; }
    TY   OP^    (CEP&c)          { return a[0]*c[1] - a[1]*c[0]; }
    EPA& OP=    (CTY*b)          { FT(a,b,=)    return *this; }
    EPA& OP+=   (CTY*b)          { FT(a,b,+=)   return *this; }
    EPA& OP-=   (CTY*b)          { FT(a,b,-=)   return *this; }
    PAI  OP+    (CTY*b)    const { PAI y(*a); y+=b; return y; }
    PAI  OP-    (CTY*b)    const { PAI y(*a); y-=b; return y; }
    void negate ()               { FT(a,a,=-) }
    TY   norm   ()         const { return a[0]*a[0]+a[1]*a[1]; }
    TY   OP*    (CPA &c)   const { return a[0]*c[0]+a[1]*c[1]; }
    TY   OP*    (CEP &c)   const { return a[0]*c.a[0]+a.a[1]*c.a[1]; }
    bool OP==   (CEP &c)   const { if(a[0]!=c.a[0] || a[1]!=c.a[1])
                                     return false;
                                   return true; }
    bool OP==   (CPA &c)   const { if(a[0]!=c[0] || a[1]!=c[1])
                                     return false;
                                   return true; }
    bool OP==   (CTY  x)   const { if(a[0]!=x || a[1]!=x)
                                     return false;
                                   return true; }
    bool OP!=   (CEP &c)   const { return !(*this==c); }
    bool OP!=   (CPA &c)   const { return !(*this==c); }
    bool OP!=   (CTY  x)   const { return !(*this==x); }
    EPA& add_mul(CEP& c, CTY x)  { a[0]+=x*c[0]; a[1]+=x*c[1]; return *this; }
    EPA& add_mul(CPA& c, CTY x)  { a[0]+=x*c[0]; a[1]+=x*c[1]; return *this; }
  };

#undef FT
#undef FTX

  TM TY   OP*    (CPA    &a, CEP<TY>&b) { return b*a; } 
  TM PAI& OP+=   (PAI    &a, CEP<TY>&b) { return a+=(CTY*)(b); }
  TM PAI& OP-=   (PAI    &a, CEP<TY>&b) { return a-=(CTY*)(b); }
  TM PAI  OP+    (CPA    &a, CEP<TY>&b) { return a+ (CTY*)(b); }
  TM PAI  OP-    (CPA    &a, CEP<TY>&b) { return a- (CTY*)(b); }
  TM TY   dist_sq(CPA    &a, CEP<TY>&b) { return   square(a[0]-b[0])
				                 + square(a[1]-b[1]); }
  TM TY   dist_sq(CEP<TY>&a, CPA    &b) { return   square(a[0]-b[0])
				                 + square(a[1]-b[1]); }
  TM TY   dist_sq(CEP<TY>&a, CEP<TY>&b) { return   square(a[0]-b[0])
				                 + square(a[1]-b[1]); }
  TM TY   sum_sq (CPA    &a, CEP<TY>&b) { return   square(a[0]+b[0])
				                 + square(a[1]+b[1]); }
  TM TY   sum_sq (CEP<TY>&a, CPA    &b) { return   square(a[0]+b[0])
				                 + square(a[1]+b[1]); }
  TM TY   sum_sq (CEP<TY>&a, CEP<TY>&b) { return   square(a[0]+b[0])
				                 + square(a[1]+b[1]); }

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::esym22 for a symmetric 2x2 matrix using external memory    
  //////////////////////////////////////////////////////////////////////////////

#define EYM         esym22
#define CEY         const esym22
#define FI          for(register unsigned short i=0; i<3; i++)
#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;


  template<class TY>
  class EYM {
  private:
    TY *a;
    //   element mapping is as follows: 0 -> 0,0
    //                                  1 -> 0,1 & 1,0
    //                                  2 -> 1,1
    EYM          (CEY &);                                      // disabled!
  public:
    EYM          ( TY*b)          { a=b; }
    EYM& unity   ()               { a[0]=a[2]=TY(1); a[1]=TY(0); return *this; }
         OP TY* ()                { return a; }
         OP CTY*() const          { return a; }
    EYM& OP=     (CTY x)          { FTX(=)     return *this; }
    EYM& OP*=    (CTY x)          { FTX(*=)    return *this; }
    EYM& OP/=    (CTY x)          { FTX(/=)    return *this; }
    EYM& OP=     (CSY&c)          { FT(a,c,=)  return *this; }
    EYM& OP+=    (CSY&c)          { FT(a,c,+=) return *this; }
    EYM& OP-=    (CSY&c)          { FT(a,c,-=) return *this; }
    EYM& OP=     (CEY&c)          { FT(a,c.a,=)  return *this; }
    EYM& OP+=    (CEY&c)          { FT(a,c.a,+=) return *this; }
    EYM& OP-=    (CEY&c)          { FT(a,c.a,-=) return *this; }
    CTY& OP[]    (CI  i)    const { return a[i]; }
    CTY& OP[]    (CUI i)    const { return a[i]; }
    CTY& OP[]    (CUS i)    const { return a[i]; }
    TY&  OP[]    (CI  i)          { return a[i]; }
    TY&  OP[]    (CUI i)          { return a[i]; }
    TY&  OP[]    (CUS i)          { return a[i]; }
    CTY& OP()    (CI  i, CI  j) const { return a[i+j]; }
    CTY& OP()    (CUI i, CUI j) const { return a[i+j]; }
    CTY& OP()    (CUS i, CUS j) const { return a[i+j]; }
    TY&  OP()    (CI  i, CI  j)       { return a[i+j]; }
    TY&  OP()    (CUI i, CUI j)       { return a[i+j]; }
    TY&  OP()    (CUS i, CUS j)       { return a[i+j]; }
    void negate  ()               { FT(a,a,=-) }
    TY   trace   ()         const { return a[0] + a[2]; }
    bool OP==    (CSY&c)    const { if(a[0]!=c[0] || a[1]!=c[1] || a[2]!=c[2]) 
                                      return false;
                                    return true; }
    bool OP==    (CEY&c)    const { if(a[0]!=c.a[0] ||
				       a[1]!=c.a[1] ||
				       a[2]!=c.a[2]) return false;
                                    return true; }
    bool OP!=    (CSY&c)    const { return !(*this==c); }
    bool OP!=    (CEY&c)    const { return !(*this==c); }
    PAI  OP*     (CPA& x)   const
      // x_i * A_ij
    {
      register TY w[2] = { a[0]*x[0]+a[1]*x[1], a[1]*x[0]+a[2]*x[1]}; 
      return PAI(w);
    }
    TY   contract(CPA& x) const
      // Xi*Xj*Aij
    {
      return x[0] * (a[0]*x[0] + a[1]*x[1]) +
	     x[1] * (a[1]*x[0] + a[2]*x[1]);
    }
    TY   contract(CSY& M) const
      // Aij*Bij
    {
      register TY x =     a[1]*M[1];
      return      x + x + a[0]*M[0] + a[2]*M[2];
    }
    TY   contract(CEY& M) const
      // Aij*Bij
    {
      register TY x =     a[1]*M.a[1];
      return      x + x + a[0]*M.a[0] + a[2]*M.a[2];
    }
    EYM& outprod (CPA& x)
      // Xi*Xj
    {
      a[0]=x[0]*x[0];
      a[1]=x[0]*x[1];
      a[2]=x[1]*x[1];
      return *this;
    }
    EYM& outprod (CPA& x, CPA& y)
      // Xi*Xj + Yi*Xj
    {
      a[0] =x[0]*y[0];
      a[0]+=a[0];
      a[1] =x[0]*y[1] + x[1]*y[0];
      a[2] =x[1]*y[1];
      a[2]+=a[2];
      return *this;
    }
    EYM& unity (CTY A)
      // a*Dij
    {
      a[0]=a[2]= A;
      a[1]=TY(0);
      return *this;
    }
    EYM& add_unity (CTY A)
      // add a*Dij
    {
      a[0]+=A;
      a[2]+=A;
      return *this;
    }
    EYM& quadrup (CPA& x, CTY A, CTY B)
      // a*Dij + b*Xi*Xj
    {
      register TY Bx0 = B*x[0];
      a[0]=A+Bx0*x[0];
      a[1]=Bx0*x[1];
      a[2]=A+B*x[1]*x[1];
      return *this;
    }
    EYM& add_quadrup (CPA& x, CTY A, CTY B)
      // a*Dij + b*Xi*Xj
    {
      register TY Bx0 = B*x[0];
      a[0]+=A+Bx0*x[0];
      a[1]+=Bx0*x[1];
      a[2]+=A+B*x[1]*x[1];
      return *this;
    }
    EYM& add_mul (CEY& c, CTY x)
    {
      FI a[i] += x*c.a[i];
      return *this;
    }
    EYM& add_mul (CSY& c, CTY x)
    {
      FI a[i] += x*c[i];
      return *this;
    }
  };

#undef  FTX
#undef  FT
#undef  FI
#undef  OPIJ

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::esym222 for an symmetric 2x2x2 tensor using external memory
  //////////////////////////////////////////////////////////////////////////////
#define EYT     esym222
#define CET     const esym222

#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];  a[3] op b[3];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;     a[3] op x; 


  template<class TY> class EYT {
  private:
    TY *a;
    // element mapping is as follows: 0 -> 0,0,0
    //                                1 -> 0,0,1
    //                                2 -> 0,1,1
    //                                3 -> 1,1,1
    EYT          (CET &);                                      // disabled!
  public:
    EYT             ( TY*b)        { a=b; }
    EYT& unity   ()                { a[0]=a[3]=TY(1); a[1]=a[2]=TY(0);
                                     return *this; }
          OP TY*  ()               { return a; }
          OP CTY* () const         { return a; }
    EYT&  OP=     (CTY x)          { FTX(=)  return *this; }
    EYT&  OP*=    (CTY x)          { FTX(*=) return *this; }
    EYT&  OP/=    (CTY x)          { FTX(/=) return *this; }
    EYT&  OP=     (CST&c)          { FT(a,c,=)  return *this; }
    EYT&  OP+=    (CST&c)          { FT(a,c,+=) return *this; }
    EYT&  OP-=    (CST&c)          { FT(a,c,-=) return *this; }
    EYT&  OP=     (CET&c)          { FT(a,c.a,=)  return *this; }
    EYT&  OP+=    (CET&c)          { FT(a,c.a,+=) return *this; }
    EYT&  OP-=    (CET&c)          { FT(a,c.a,-=) return *this; }
    CTY&  OP[]    (CI  i)    const { return a[i]; }
    CTY&  OP[]    (CUI i)    const { return a[i]; }
    CTY&  OP[]    (CUS i)    const { return a[i]; }
    TY&   OP[]    (CI  i)          { return a[i]; }
    TY&   OP[]    (CUI i)          { return a[i]; }
    TY&   OP[]    (CUS i)          { return a[i]; }
    CTY&  OP()    (CI  i, CI  j, CI  k) const { return a[i+j+k]; }
    CTY&  OP()    (CUI i, CUI j, CUI k) const { return a[i+j+k]; }
    CTY&  OP()    (CUS i, CUS j, CUS k) const { return a[i+j+k]; }
    TY&   OP()    (CI  i, CI  j, CI  k)       { return a[i+j+k]; }
    TY&   OP()    (CUI i, CUI j, CUI k)       { return a[i+j+k]; }
    TY&   OP()    (CUS i, CUS j, CUS k)       { return a[i+j+k]; }
    void  negate  ()               { FT(a,a,=-) }
    bool  OP==    (CST&c)    const { if(a[0]!=c[0] || a[1]!=c[1] ||
					a[2]!=c[2] || a[3]!=c[3] ) return false;
   				     return true; }
    bool  OP==    (CET&c)    const { if(a[0]!=c.a[0] || a[1]!=c.a[1] ||
					a[2]!=c.a[2] || a[3]!=c.a[3] )
    				       return false;
    				     return true; }
    bool  OP!=    (CST&c)    const { return !(*this==c); }
    bool  OP!=    (CET&c)    const { return !(*this==c); }
    PAI   trace   ()         const
      // P_ijk*D_jk
    {
      register TY w[3] = { a[0]+a[2], a[1]+a[3]};
      return PAI(w);
    }
    TY    trace   (CPA& x)   const
      // trace(x*A)
    {
      return   x[0]*(a[0]+a[2]) + x[1]*(a[1]+a[3]);
    }
    PAI   OP*     (CSY& M)   const
      // A_ijk*Mij
    {
      register PAI W;
      W[0] = a[1]*M[1];
      W[0]+= W[0] + a[0]*M[0] + a[2]*M[2];
      W[1] = a[2]*M[1];
      W[1]+= W[1] + a[1]*M[0] + a[3]*M[2];
      return W;
    }
    PAI   OP*     (CEY<TY>& M) const
      // A_ijk*Mij
    {
      register PAI W;
      W[0] = a[1]*M[1];
      W[0]+= W[0] + a[0]*M[0] + a[2]*M[2];
      W[1] = a[2]*M[1];
      W[1]+= W[1] + a[1]*M[0] + a[3]*M[2];
      return W;
    }
    SYM   OP*     (CPA& x) const
      // A_ijk*Xi
    {
      register SYM M;
      M[0] =a[0]*x[0] +a[1]*x[1];
      M[1] =a[1]*x[0] +a[2]*x[1];
      M[2] =a[2]*x[0] +a[3]*x[1];
      return M;
    }
    EYT&    outprod (CPA& x)
      // Xi*Xj*Xk
    {
      register TY tmp=x[0]*x[0];
      a[0]= tmp*x[0];
      a[1]= tmp*x[1];
      tmp = x[1]*x[1];
      a[2]= tmp*x[0];
      a[3]= tmp*x[1];
      return *this;
    }
    EYT&    outprod (CPA& x, CSY& M)
      // Xi*Mjk +Xj*Mik +Xk*Mij
    {
      a[0] =M[0]*x[0];
      a[0]+=a[0]+a[0];
      a[1] =M[1]*x[0];
      a[1]+=a[1]+M[0]*x[1];
      a[2] =M[1]*x[1];
      a[2]+=a[1]+M[2]*x[0];
      a[3] =M[2]*x[1];
      a[3]+=a[3]+a[3];
      return *this;
    } 
    EYT&    outprod (CPA& x, CEY<TY>& M)
      // Xi*Mjk +Xj*Mik +Xk*Mij
    {
      a[0] =M[0]*x[0];
      a[0]+=a[0]+a[0];
      a[1] =M[1]*x[0];
      a[1]+=a[1]+M[0]*x[1];
      a[2] =M[1]*x[1];
      a[2]+=a[1]+M[2]*x[0];
      a[3] =M[2]*x[1];
      a[3]+=a[3]+a[3];
      return *this;
    } 
    EYT&    quadrup (CPA& x, CTY A)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi);
    {
      register TY ai=A*x[0];
      a[0]=ai+ai+ai;
      a[2]=ai;
      ai  =A*x[1];
      a[3]=ai+ai+ai;
      a[1]=ai;
      return *this;
    }
    EYT&    add_quadrup (CPA& x, CTY A)
      // Pijk += a (Dij Xk + Dik Xj + Djk Xi);
    {
      register TY ai=A*x[0];
      a[0]+=ai+ai+ai;
      a[2]+=ai;
      ai   =A*x[1];
      a[3]+=ai+ai+ai;
      a[1]+=ai;
      return *this;
    }
    EYT&    quadrup (CPA& x, CTY A, CTY B)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi) + b Xi Xj Xk;
    {
      register TY A3=A+A+A, tmp=B*x[0]*x[0];
      a[0] =(A3+tmp)*x[0];
      a[1] =(A +tmp)*x[1];
      tmp  =B*x[1]*x[1];
      a[2] =(A +tmp)*x[0]; 
      a[3] =(A3+tmp)*x[1];
      return *this;
    }
    EYT&    add_quadrup (CPA& x, CTY A, CTY B)
      // Pijk += a (Dij Xk + Dik Xj + Djk Xi) + b Xi Xj Xk;
    {
      register TY A3=A+A+A, tmp=B*x[0]*x[0];
      a[0]+=(A3+tmp)*x[0];
      a[1]+=(A +tmp)*x[1];
      tmp  =B*x[1]*x[1];
      a[2]+=(A +tmp)*x[0]; 
      a[3]+=(A3+tmp)*x[1];
      return *this;
    }
  };

#undef  FTX
#undef  FT
#undef  OPIJK

  ////////////////////////////////////////////////////////////////////////////////
  // more
  ////////////////////////////////////////////////////////////////////////////////

  TM bool OP==(CPA &a, CEP<TY>&b) { return b==a; }
  TM bool OP!=(CPA &a, CEP<TY>&b) { return b!=a; }

#ifndef C_AND_F
  TM std::ostream& OP<<(std::ostream& s, CEP<TY>&a) { return s<<a[0]<<' '<<a[1]; }
  TM std::istream& OP>>(std::istream& s, EPA<TY>&a) {
    TY x[2];
    char c=0;    s>>c;
    if(c == '(') { s>>x[0]>>x[1]; s>>c; if(c!=')') s.clear(std::ios::badbit); }
    else         { s.putback(c); s>>x[0]>>x[1]; }
    if(s) a=x;
    return s;
  }
  TM std::ostream& OP<< (std::ostream& s, CEY<TY>&M)
  { return s<<M[0]<<' '<<M[1]<<' '<<M[2]; }

  TM std::ostream& OP<< (std::ostream& s, CET<TY>&N) 
  {  return s<<N[0]<<' '<<N[1]<<' '<<N[2]<<' '<<N[3]; }
#endif // C_AND_F

  TM PAI OP*  (CPA& X, CEY<TY>& M) { return M*X; }
  TM SYM OP*  (CPA& X, CET<TY>& N) { return N*X; }
  TM TY  trace(CEY<TY>& M)         { return M.trace(); }
  TM TY  trace(CET<TY>& M, CPA& x) { return M.trace(x); }
  TM PAI trace(CET<TY>& M)         { return M.trace(); }

}
////////////////////////////////////////////////////////////////////////////////
#undef  CTY
#undef  PAI
#undef  CPA
#undef  EPA
#undef  CEP
#undef  SYM
#undef  CSY
#undef  SYT
#undef  CST
#undef  EYM
#undef  CEY
#undef  EYT
#undef  CET
#undef  CI
#undef  CUI
#undef  CUS
#undef  TM
#undef  OP
////////////////////////////////////////////////////////////////////////////////
#endif  // included_et2D_h
