// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// et3D.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class TY> class etripel                fully inline                |
// template<class TY> class esym33                 fully inline                |
// template<class TY> class esym333                fully inline                |
// some global functions                           fully inline                |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_et3D_h
#define included_et3D_h

#ifndef included_tn3D_h
#include <tree/tn3D.h>
#endif

////////////////////////////////////////////////////////////////////////////////

#define CTY         const TY
#define CI          const int
#define CUI         const unsigned int
#define CUS         const unsigned short
#define TRI         tripel<TY>
#define CTR         const tripel<TY>
#define SYM         sym33<TY>
#define CSY         const sym33<TY>
#define SYT         sym333<TY>
#define CST         const sym333<TY>
#define OP          operator
#define TM          template<class TY> inline

namespace fasttree {

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::etripel for a tripel using external memory                 
  //////////////////////////////////////////////////////////////////////////////

#define ETR         etripel
#define CER         const etripel

#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;

  template<class TY>
  class ETR {
  private:
    TY *a;
    ETR         ();                                           // disabled!
//     ETR         (CER &);                                      // disabled!
  public:
    CTY& OP[]   (CI  i)    const { return a[i]; }
    CTY& OP[]   (CUI i)    const { return a[i]; }
    CTY& OP[]   (CUS i)    const { return a[i]; }
    TY&  OP[]   (CI  i)          { return a[i]; }
    TY&  OP[]   (CUI i)          { return a[i]; }
    TY&  OP[]   (CUS i)          { return a[i]; }
    ETR         (TY*b)           { a=b; }
         OP TY* ()               { return a; }
         OP CTY*()         const { return a; }
    ETR& OP=    (CTY x)          { FTX(=)     return *this; }
    ETR& OP*=   (CTY x)          { FTX(*=)    return *this; }
    ETR& OP/=   (CTY x)          { FTX(/=)    return *this; }
    TRI  OP*    (CTY x)    const { TRI y((CTY*)(*this)); y*=x; return y; }
    TRI  OP/    (CTY x)    const { TRI y((CTY*)(*this)); y/=x; return y; }
    ETR& OP=    (CTR&c)          { FT(a,c,=)  return *this; }
    ETR& OP+=   (CTR&c)          { FT(a,c,+=) return *this; }
    ETR& OP-=   (CTR&c)          { FT(a,c,-=) return *this; }
    TRI  OP+    (CTR&c)    const { TRI y(a); y+=c; return y; }
    TRI  OP-    (CTR&c)    const { TRI y(a); y-=c; return y; }
    TRI  OP^    (CTR&c)    const { TRI z;
                                   z[0] = a[1]*c[2] - a[2]*c[1];
			   	   z[1] = a[2]*c[0] - a[0]*c[2];
			           z[2] = a[0]*c[1] - a[1]*c[0];
				   return z; }
    ETR& OP=    (CER&c)          { FT(a,c.a,=)  return *this; }
    ETR& OP+=   (CER&c)          { FT(a,c.a,+=) return *this; }
    ETR& OP-=   (CER&c)          { FT(a,c.a,-=) return *this; }
    TRI  OP+    (CER&c)    const { TRI y(*a); y+=c.a; return y; }
    TRI  OP-    (CER&c)    const { TRI y(*a); y-=c.a; return y; }
    TRI  OP^    (CER&c)    const { TRI z;
                                   z[0] = a[1]*c[2] - a[2]*c[1];
			   	   z[1] = a[2]*c[0] - a[0]*c[2];
			           z[2] = a[0]*c[1] - a[1]*c[0];
				   return z; }
    ETR& OP=    (CTY*b)          { FT(a,b,=)    return *this; }
    ETR& OP+=   (CTY*b)          { FT(a,b,+=)   return *this; }
    ETR& OP-=   (CTY*b)          { FT(a,b,-=)   return *this; }
    TRI  OP+    (CTY*b)    const { TRI y(*a); y+=b; return y; }
    TRI  OP-    (CTY*b)    const { TRI y(*a); y-=b; return y; }
    void negate ()               { FT(a,a,=-) }
    TY   norm   ()         const { return a[0]*a[0]+a[1]*a[1]+a[2]*a[2]; }
    TY   OP*    (CTR &c)   const { return a[0]*c[0]+a[1]*c[1]+a[2]*c[2]; }
    TY   OP*    (CER &c)   const { return a[0]*c.a[0]+a.a[1]*c.a[1]+a[2]*c.a[2]; }
    bool OP==   (CER &c)   const { if(a[0]!=c.a[0] ||
				      a[1]!=c.a[1] ||
				      a[2]!=c.a[2]) return false;
                                   return true; }
    bool OP==   (CTR &c)   const { if(a[0]!=c[0] || a[1]!=c[1] || a[2]!=c[2])
                                     return false;
                                   return true; }
    bool OP==   (CTY  x)   const { if(a[0]!=x || a[1]!=x || a[2]!=x)
                                     return false;
                                   return true; }
    bool OP!=   (CER &c)   const { return !(*this==c); }
    bool OP!=   (CTR &c)   const { return !(*this==c); }
    bool OP!=   (CTY  x)   const { return !(*this==x); }
    ETR& add_mul (CER& c, CTY x) { a[0]+=x*c[0]; a[1]+=x*c[1]; a[2]+=x*c[2];
                                   return *this; }
    ETR& add_mul (CTR& c, CTY x) { a[0]+=x*c[0]; a[1]+=x*c[1]; a[2]+=x*c[2];
                                   return *this; }
  };

#undef FT
#undef FTX

  TM TY   OP*    (CTR    &a, CER<TY>&b) { return b*a; } 
  TM TRI& OP+=   (TRI    &a, CER<TY>&b) { return a+=(CTY*)(b); }
  TM TRI& OP-=   (TRI    &a, CER<TY>&b) { return a-=(CTY*)(b); }
  TM TRI  OP+    (CTR    &a, CER<TY>&b) { return a+ (CTY*)(b); }
  TM TRI  OP-    (CTR    &a, CER<TY>&b) { return a- (CTY*)(b); }
  TM TY   dist_sq(CTR    &a, CER<TY>&b) { return   square(a[0]-b[0])
				                 + square(a[1]-b[1])
				                 + square(a[2]-b[2]); }
  TM TY   dist_sq(CER<TY>&a, CTR    &b) { return   square(a[0]-b[0])
				                 + square(a[1]-b[1])
				                 + square(a[2]-b[2]); }
  TM TY   dist_sq(CER<TY>&a, CER<TY>&b) { return   square(a[0]-b[0])
				                 + square(a[1]-b[1])
				                 + square(a[2]-b[2]); }
  TM TY   sum_sq (CTR    &a, CER<TY>&b) { return   square(a[0]+b[0])
				                 + square(a[1]+b[1])
				                 + square(a[2]+b[2]); }
  TM TY   sum_sq (CER<TY>&a, CTR    &b) { return   square(a[0]+b[0])
				                 + square(a[1]+b[1])
				                 + square(a[2]+b[2]); }
  TM TY   sum_sq (CER<TY>&a, CER<TY>&b) { return   square(a[0]+b[0])
				                 + square(a[1]+b[1])
				                 + square(a[2]+b[2]); }

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::esym33 for a symmetric 3x3 matrix using external memory    
  //////////////////////////////////////////////////////////////////////////////

#define EYM         esym33
#define CEY         const esym33
#define FI          for(register unsigned short i=0; i<6; i++)
#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];			\
                    a[3] op b[3]; a[4] op b[4]; a[5] op b[5];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;			\
                    a[3] op x;    a[4] op x;    a[5] op x;
#define OPIJ        if(i>j) return (*this)(j,i);				\
                    switch(i+i+j) { 						\
                    case 0: return a[0];					\
                    case 1: return a[1];					\
                    case 2: return a[2];					\
                    case 3: return a[3];					\
                    case 4: return a[4];					\
                    default:return a[5]; }


  template<class TY>
  class EYM {
  private:
    TY *a;
    //   element mapping is as follows: 0 -> 0,0
    //                                  1 -> 0,1 & 1,0
    //                                  2 -> 0,2 & 2,0
    //                                  3 -> 1,1
    //                                  4 -> 1,2 & 2,1
    //                                  5 -> 2,2 & 2,2
    EYM          ();                                           // disabled!
//     EYM          (CEY &);                                      // disabled!
  public:
    EYM          (TY*b)           { a=b; }
    EYM& unity   ()               { a[0]=a[3]=a[5]=TY(1); a[1]=a[2]=a[4]=TY(0);
                                    return *this; }
         OP TY*  ()               { return a; }
         OP CTY* () const         { return a; }
    EYM& OP=     (CTY x)          { FTX(=)     return *this; }
    EYM& OP*=    (CTY x)          { FTX(*=)    return *this; }
    EYM& OP/=    (CTY x)          { FTX(/=)    return *this; }
    EYM& OP=     (CSY&c)          { FT(a,c,=)  return *this; }
    EYM& OP+=    (CSY&c)          { FT(a,c,+=) return *this; }
    EYM& OP-=    (CSY&c)          { FT(a,c,-=) return *this; }
    EYM& OP=     (CEY&c)          { FT(a,c.a,=)  return *this; }
    EYM& OP+=    (CEY&c)          { FT(a,c.a,+=) return *this; }
    EYM& OP-=    (CEY&c)          { FT(a,c.a,-=) return *this; }
    CTY& OP[]    (CI  i)    const { return a[i]; }
    CTY& OP[]    (CUI i)    const { return a[i]; }
    CTY& OP[]    (CUS i)    const { return a[i]; }
    TY&  OP[]    (CI  i)          { return a[i]; }
    TY&  OP[]    (CUI i)          { return a[i]; }
    TY&  OP[]    (CUS i)          { return a[i]; }
    CTY& OP()    (CI  i, CI  j) const { OPIJ }
    CTY& OP()    (CUI i, CUI j) const { OPIJ }
    CTY& OP()    (CUS i, CUS j) const { OPIJ }
    TY&  OP()    (CI  i, CI  j)       { OPIJ }
    TY&  OP()    (CUI i, CUI j)       { OPIJ }
    TY&  OP()    (CUS i, CUS j)       { OPIJ }
    void negate  ()               { FT(a,a,=-) }
    TY   trace   ()         const { return a[0] + a[3] + a[5]; }
    bool OP==    (CSY&c)    const { if(a[0]!=c[0] || a[1]!=c[1] ||
				       a[2]!=c[2] || a[3]!=c[3] ||
				       a[4]!=c[4] || a[5]!=c[5] ) return false;
                                    return true; }
    bool OP==    (CEY&c)    const { if(a[0]!=c.a[0] || a[1]!=c.a[1] ||
				       a[2]!=c.a[2] || a[3]!=c.a[3] ||
				       a[4]!=c.a[4] || a[5]!=c.a[5]) return false;
                                    return true; }
    bool OP!=    (CSY&c)    const { return !(*this==c); }
    bool OP!=    (CEY&c)    const { return !(*this==c); }
    TRI  OP*     (CTR&x)    const 
      // x_i * A_ij
    {
      register TY w[3] = { a[0]*x[0] + a[1]*x[1] + a[2]*x[2],
			   a[1]*x[0] + a[3]*x[1] + a[4]*x[2], 
			   a[2]*x[0] + a[4]*x[1] + a[5]*x[2] }; 
      return TRI(w);
    }
    TY   contract(CTR& x) const
      // Xi*Xj*Aij
    {
      return x[0] * (a[0]*x[0] + a[1]*x[1] + a[2]*x[2]) +
	     x[1] * (a[1]*x[0] + a[3]*x[1] + a[4]*x[2]) + 
	     x[2] * (a[2]*x[0] + a[4]*x[1] + a[5]*x[2]);
    }
    TY   contract(CSY& M) const
      // Aij*Bij
    {
      register TY x =     a[1]*M[1] + a[2]*M[2] + a[4]*M[4];
      return      x + x + a[0]*M[0] + a[3]*M[3] + a[5]*M[5];
    }
    TY   contract(CEY& M) const
      // Aij*Bij
    {
      register TY x =     a[1]*M.a[1] + a[2]*M.a[2] + a[4]*M.a[4];
      return      x + x + a[0]*M.a[0] + a[3]*M.a[3] + a[5]*M.a[5];
    }
    EYM& outprod (CTR& x)
      // Xi*Xj
    {
      a[0]=x[0]*x[0];
      a[1]=x[0]*x[1];
      a[2]=x[0]*x[2]; 
      a[3]=x[1]*x[1];
      a[4]=x[1]*x[2];
      a[5]=x[2]*x[2];
      return *this;
    }
    EYM& outprod (CTR& x, CTR& y)
      // Xi*Xj + Yi*Xj
    {
      a[0] =x[0]*y[0];
      a[0]+=a[0];
      a[1] =x[0]*y[1] + x[1]*y[0];
      a[2] =x[0]*y[2] + x[2]*y[0];
      a[3] =x[1]*y[1];
      a[3]+=a[3];
      a[4] =x[1]*y[2] + x[2]*y[1];
      a[5] =x[2]*y[2];
      a[5]+=a[5];
      return *this;
    }
    EYM& unity (CTY A)
      // a*Dij
    {
      a[0]=a[3]=a[5] = A;
      a[1]=a[2]=a[4] = TY(0);
      return *this;
    }
    EYM& add_unity (CTY A)
      // add a*Dij
    {
      a[0]+=A;
      a[3]+=A;
      a[5]+=A;
      return *this;
    }
    EYM& quadrup (CTR& x, CTY A, CTY B)
      // a*Dij + b*Xi*Xj
    {
      register TY tmp=B*x[0];
      a[0]=A+tmp*x[0];
      a[1]=  tmp*x[1];
      a[2]=  tmp*x[2];
      tmp =B*x[1];
      a[3]=A+tmp*x[1];
      a[4]=  tmp*x[2];
      a[5]=A+B*x[2]*x[2];
      return *this;
    }
    EYM& add_quadrup (CTR& x, CTY A, CTY B)
      // a*Dij + b*Xi*Xj
    {
      register TY tmp=B*x[0];
      a[0]+=A+tmp*x[0];
      a[1]+=  tmp*x[1];
      a[2]+=  tmp*x[2];
      tmp  =B*x[1];
      a[3]+=A+tmp*x[1];
      a[4]+=tmp*x[2];
      a[5]+=A+B*x[2]*x[2];
      return *this;
    }
    EYM& add_mul (CEY& c, CTY x)
    {
      FI a[i] += x*c.a[i];
      return *this;
    }
    EYM& add_mul (CSY& c, CTY x)
    {
      FI a[i] += x*c[i];
      return *this;
    }
  };

#undef  FTX
#undef  FT
#undef  FI
#undef  OPIJ

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::esym333 for an symmetric 3x3x3 tensor using external memory
  //////////////////////////////////////////////////////////////////////////////

#define EYT     esym333
#define CET     const esym333

#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];			\
                    a[3] op b[3]; a[4] op b[4]; a[5] op b[5];			\
                    a[6] op b[6]; a[7] op b[7]; a[8] op b[8]; a[9] op b[9];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;			\
                    a[3] op x;    a[4] op x;    a[5] op x;			\
                    a[6] op x;    a[7] op x;    a[8] op x;    a[9] op x;
#define OPIJK	    if(i>j)        { if(j>k) return (*this)(k,j,i);		\
                                     if(i>k) return (*this)(j,k,i);		\
                                        else return (*this)(j,i,k);		\
                    } else if(j>k) { if(i>k) return (*this)(k,i,j);		\
                                        else return (*this)(i,k,j);		\
                    } else { switch(i+i+i+i+j+j+k) {				\
                             case  0: return a[0];				\
                             case  1: return a[1];				\
                             case  2: return a[2];				\
                             case  3: return a[3];				\
                             case  4: return a[4];				\
                             case  6: return a[5];				\
                             case  7: return a[6];				\
                             case  8: return a[7];				\
                             case  9: return a[8];				\
                             default: return a[9]; } }

  template<class TY>
  class EYT {
  private:
    TY *a;
    // element mapping is as follows: 0 -> 0,0,0
    //                                1 -> 0,0,1
    //                                2 -> 0,0,2
    //                                3 -> 0,1,1
    //                                4 -> 0,1,2
    //                                5 -> 0,2,2
    //                                6 -> 1,1,1
    //                                7 -> 1,1,2
    //                                8 -> 1,2,2
    //                                9 -> 2,2,2
    EYT           ();                                      // disabled!
//     EYT           (CET &);                                 // disabled!
  public:
    EYT           ( TY*b)          { a=b; }
    EYT&  unity   ()               { a[1]=a[2]=a[3]=a[4]=a[5]=a[7]=a[8]=TY(0);
                                     a[0]=a[6]=a[9]=TY(1); return *this; }
          OP TY*  ()               { return a; }
          OP CTY* () const         { return a; }
    EYT&  OP=     (CTY x)          { FTX(=)  return *this; }
    EYT&  OP*=    (CTY x)          { FTX(*=) return *this; }
    EYT&  OP/=    (CTY x)          { FTX(/=) return *this; }
    EYT&  OP=     (CST&c)          { FT(a,c,=)  return *this; }
    EYT&  OP+=    (CST&c)          { FT(a,c,+=) return *this; }
    EYT&  OP-=    (CST&c)          { FT(a,c,-=) return *this; }
    EYT&  OP=     (CET&c)          { FT(a,c.a,=)  return *this; }
    EYT&  OP+=    (CET&c)          { FT(a,c.a,+=) return *this; }
    EYT&  OP-=    (CET&c)          { FT(a,c.a,-=) return *this; }
    CTY&  OP[]    (CI  i)    const { return a[i]; }
    CTY&  OP[]    (CUI i)    const { return a[i]; }
    CTY&  OP[]    (CUS i)    const { return a[i]; }
    TY&   OP[]    (CI  i)          { return a[i]; }
    TY&   OP[]    (CUI i)          { return a[i]; }
    TY&   OP[]    (CUS i)          { return a[i]; }
    CTY&  OP()    (CI  i, CI  j, CI  k) const { OPIJK }
    CTY&  OP()    (CUI i, CUI j, CUI k) const { OPIJK }
    CTY&  OP()    (CUS i, CUS j, CUS k) const { OPIJK }
    TY&   OP()    (CI  i, CI  j, CI  k)       { OPIJK }
    TY&   OP()    (CUI i, CUI j, CUI k)       { OPIJK }
    TY&   OP()    (CUS i, CUS j, CUS k)       { OPIJK }
    void  negate  ()               { FT(a,a,=-) }
    bool  OP==    (CST&c)    const { if(a[0]!=c[0] || a[1]!=c[1] ||
					a[2]!=c[2] || a[3]!=c[3] ||
					a[4]!=c[4] || a[5]!=c[5] ||
					a[6]!=c[6] || a[7]!=c[7] ||
					a[8]!=c[8] || a[9]!=c[9] ) return false;
                                     return true; }
    bool  OP==    (CET&c)    const { if(a[0]!=c.a[0] || a[1]!=c.a[1] ||
					a[2]!=c.a[2] || a[3]!=c.a[3] ||
					a[4]!=c.a[4] || a[5]!=c.a[5] ||
					a[6]!=c.a[6] || a[7]!=c.a[7] ||
					a[8]!=c.a[8] || a[9]!=c.a[9] )
                                       return false;
                                     return true; }
    bool  OP!=    (CST&c)    const { return !(*this==c); }
    bool  OP!=    (CET&c)    const { return !(*this==c); }
    TRI   trace   ()         const
      // P_ijk*D_jk
    {
      register TY w[3] = { a[0] + a[3] + a[5],
			   a[1] + a[6] + a[8],
			   a[2] + a[7] + a[9] };
      return TRI(w);
    }
    TY    trace   (CTR& x)   const
      // trace(x*A)
    {
      return   x[0]*(a[0]+a[3]+a[5])
	     + x[1]*(a[1]+a[6]+a[8])
	     + x[2]*(a[2]+a[7]+a[9]);
    }
    TRI   OP*     (CSY& M)   const
      // A_ijk*Mij
    {
      register TRI W;
      register TY x=    a[1]*M[1]+a[2]*M[2]+a[4]*M[4];
      W[0]         =x+x+a[0]*M[0]+a[3]*M[3]+a[5]*M[5];
      x            =    a[3]*M[1]+a[4]*M[2]+a[7]*M[4];
      W[1]         =x+x+a[1]*M[0]+a[6]*M[3]+a[8]*M[5];
      x            =    a[4]*M[1]+a[5]*M[2]+a[8]*M[4];
      W[2]         =x+x+a[2]*M[0]+a[7]*M[3]+a[9]*M[5];
      return W;
    }
    TRI   OP*     (CEY<TY>& M) const
      // A_ijk*Mij
    {
      register TRI W;
      register TY x=    a[1]*M[1]+a[2]*M[2]+a[4]*M[4];
      W[0]         =x+x+a[0]*M[0]+a[3]*M[3]+a[5]*M[5];
      x            =    a[3]*M[1]+a[4]*M[2]+a[7]*M[4];
      W[1]         =x+x+a[1]*M[0]+a[6]*M[3]+a[8]*M[5];
      x            =    a[4]*M[1]+a[5]*M[2]+a[8]*M[4];
      W[2]         =x+x+a[2]*M[0]+a[7]*M[3]+a[9]*M[5];
      return W;
    }
    SYM   OP*     (CTR& x) const
      // A_ijk*Xi
    {
      register SYM M;
      M[0] =a[0]*x[0] +a[1]*x[1] +a[2]*x[2];
      M[1] =a[1]*x[0] +a[3]*x[1] +a[4]*x[2];
      M[2] =a[2]*x[0] +a[4]*x[1] +a[5]*x[2];
      M[3] =a[3]*x[0] +a[6]*x[1] +a[7]*x[2];
      M[4] =a[4]*x[0] +a[7]*x[1] +a[8]*x[2];
      M[5] =a[5]*x[0] +a[8]*x[1] +a[9]*x[2];
      return M;
    }
    EYT&    outprod (CTR& x)
      // Xi*Xj*Xk
    {
      register TY v00=v[0]*v[0], v11=v[1]*v[1], v22=v[2]*v[2];
      a[0]= v00*v[0];
      a[1]= v00*v[1];
      a[2]= v00*v[2];
      a[3]= v11*v[0];
      a[4]= v[0]*v[1]*v[2];
      a[5]= v22*v[0];
      a[6]= v11*v[1];
      a[7]= v11*v[2];
      a[8]= v22*v[1];
      a[9]= v22*v[2];
      return *this;
    }
    EYT&    outprod (CTR& x, CSY& M)
      // Xi*Mjk +Xj*Mik +Xk*Mij
    {
      a[0] =3*M[0]*x[0];
      a[6] =3*M[3]*x[1];
      a[9] =3*M[5]*x[2];
      a[1] =M[0]*x[1] +2*M[1]*x[0];
      a[2] =M[0]*x[2] +2*M[2]*x[0];
      a[3] =M[3]*x[0] +2*M[1]*x[1];
      a[5] =M[5]*x[0] +2*M[2]*x[2];
      a[7] =M[3]*x[2] +2*M[4]*x[1];
      a[8] =M[5]*x[1] +2*M[4]*x[2];
      a[4] =M[1]*x[2] +M[2]*x[1] +M[4]*x[0];
      return *this;
    } 
    EYT&    outprod (CTR& x, CEY<TY>& M)
      // Xi*Mjk +Xj*Mik +Xk*Mij
    {
      a[0] =3*M[0]*x[0];
      a[6] =3*M[3]*x[1];
      a[9] =3*M[5]*x[2];
      a[1] =M[0]*x[1] +2*M[1]*x[0];
      a[2] =M[0]*x[2] +2*M[2]*x[0];
      a[3] =M[3]*x[0] +2*M[1]*x[1];
      a[5] =M[5]*x[0] +2*M[2]*x[2];
      a[7] =M[3]*x[2] +2*M[4]*x[1];
      a[8] =M[5]*x[1] +2*M[4]*x[2];
      a[4] =M[1]*x[2] +M[2]*x[1] +M[4]*x[0];
      return *this;
    } 
    EYT&    quadrup (CTR& x, CTY A)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi);
    {
      register TY ai=A*x[0];
      a[0]=ai+ai+ai;
      a[3]=ai;
      a[5]=ai;
      ai  =A*x[1];
      a[6]=ai+ai+ai;
      a[1]=ai;
      a[8]=ai;
      ai  =A*x[2];
      a[9]=ai+ai+ai;
      a[2]=ai;
      a[7]=ai;
      a[4]=zero;
      return *this;
    }
    EYT&    add_quadrup (CTR& x, CTY A)
      // Pijk += a (Dij Xk + Dik Xj + Djk Xi);
    {
      register TY ai=A*x[0];
      a[0]+=ai+ai+ai;
      a[3]+=ai;
      a[5]+=ai;
      ai   =A*x[1];
      a[6]+=ai+ai+ai;
      a[1]+=ai;
      a[8]+=ai;
      ai   =A*x[2];
      a[9]+=ai+ai+ai;
      a[2]+=ai;
      a[7]+=ai;
      return *this;
    }
    EYT&    quadrup (CTR& x, CTY A, CTY B)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi) + b Xi Xj Xk;
    {
      register TY A3=A+A+A, tmp=B*x[0]*x[0];
      a[0] =(A3+tmp)*x[0];
      a[1] =(A +tmp)*x[1];
      a[2] =(A +tmp)*x[2];
      tmp  =B*x[1]*x[1];
      a[3] =(A +tmp)*x[0];
      a[6] =(A3+tmp)*x[1];
      a[7] =(A +tmp)*x[2];
      tmp  =B*x[2]*x[2];
      a[5] =(A +tmp)*x[0];
      a[8] =(A +tmp)*x[1];
      a[9] =(A3+tmp)*x[2];
      a[4] = B*x[0]*x[1]*x[2];
      return *this;
    }
    EYT&    add_quadrup (CTR& x, CTY A, CTY B)
      // Pijk += a (Dij Xk + Dik Xj + Djk Xi) + b Xi Xj Xk;
    {
      register TY A3=A+A+A, tmp=B*x[0]*x[0];
      a[0]+=(A3+tmp)*x[0];
      a[1]+=(A +tmp)*x[1];
      a[2]+=(A +tmp)*x[2];
      tmp  =B*x[1]*x[1];
      a[3]+=(A +tmp)*x[0];
      a[6]+=(A3+tmp)*x[1];
      a[7]+=(A +tmp)*x[2];
      tmp  =B*x[2]*x[2];
      a[5]+=(A +tmp)*x[0];
      a[8]+=(A +tmp)*x[1];
      a[9]+=(A3+tmp)*x[2];
      a[4]+= B*x[0]*x[1]*x[2];
      return *this;
    }
  };

#undef  FTX
#undef  FT
#undef  OPIJK

  ////////////////////////////////////////////////////////////////////////////////
  // more
  ////////////////////////////////////////////////////////////////////////////////

  TM bool OP==(CTR &a, CER<TY>&b) { return b==a; }
  TM bool OP!=(CTR &a, CER<TY>&b) { return b!=a; }

#ifndef C_AND_F
  TM std::ostream& OP<<(std::ostream& s, CER<TY>&a)
  {
    return s<<a[0]<<' '<<a[1]<<' '<<a[2];
  }
  TM std::istream& OP>>(std::istream& s, ETR<TY>&a)
  {
    TY x[3];
    char c=0;    s>>c;
    if(c == '(') { s>>x[0]>>x[1]>>x[2]; s>>c; if(c!=')')
      s.clear(std::ios::badbit); }
    else         { s.putback(c); s>>x[0]>>x[1]>>x[2]; }
    if(s) a=x;
    return s;
  }
  TM std::ostream& OP<< (std::ostream& s, CEY<TY>&M)
  {
    return s<<M[0]<<' '<<M[1]<<' '<<M[2]<<' '<<M[3]<<' '<<M[4]<<' '<<M[5];
  }
  TM std::ostream& OP<< (std::ostream& s, CET<TY>&N) 
  {
    return s<<N[0]<<' '<<N[1]<<' '<<N[2]<<' '<<N[3]<<' '<<N[4]<<' '
	    <<N[5]<<' '<<N[6]<<' '<<N[7]<<' '<<N[8]<<' '<<N[9];
  }
#endif // C_AND_F

  TM TRI OP*  (CTR& X, CEY<TY>& M) { return M*X; }
  TM SYM OP*  (CTR& X, CET<TY>& N) { return N*X; }
  TM TY  trace(CEY<TY>& M)         { return M.trace(); }
  TM TY  trace(CET<TY>& M, CTR& x) { return M.trace(x); }
  TM TRI trace(CET<TY>& M)         { return M.trace(); }

}
////////////////////////////////////////////////////////////////////////////////

#undef  CTY
#undef  TRI
#undef  CTR
#undef  ETR
#undef  CER
#undef  SYM
#undef  CSY
#undef  SYT
#undef  CST
#undef  EYM
#undef  CEY
#undef  EYT
#undef  CET
#undef  CI
#undef  CUI
#undef  CUS
#undef  TM
#undef  OP
//------------------------------------------------------------------------------
#endif	// included_et3D_h
