// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// paiR.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000/2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class TY> class paiR              fully inline                     |
// some global operators                      fully inline                     |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_paiR_h
#define included_paiR_h

#ifndef included_inline_h
#include <tree/inline.h>
#endif

////////////////////////////////////////////////////////////////////////////////
// class paiR for a pair of two elements of type TY

#define CTY         const TY
#define CI          const int
#define CUI         const unsigned int
#define CUS         const unsigned short
#define PAI         paiR
#define CPA         const paiR
#define OP          operator
#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1];
#define FTX(op)     a[0] op x;    a[1] op x;

namespace fasttree {
  template<class TY>
  class PAI {
  private:
    TY a[2];
  public:
    TY&  OP[]   (CI  i)        { return a[i]; }
    CTY& OP[]   (CI  i) const  { return a[i]; }
    TY&  OP[]   (CUI i)        { return a[i]; }
    CTY& OP[]   (CUI i) const  { return a[i]; }
    TY&  OP[]   (CUS i)        { return a[i]; }
    CTY& OP[]   (CUS i) const  { return a[i]; }
    PAI         ()             {}
   ~PAI         ()             {}
    PAI         (CTY x)        { FTX(=) }
    PAI         (CTY*b)        { FT (a,b,=) }
    PAI         (CPA&c)        { FT (a,c.a,=) }
    PAI& OP=    (CTY x)        { FTX(=)  return *this; }
    PAI& OP*=   (CTY x)        { FTX(*=) return *this; }
    PAI& OP/=   (CTY x)        { FTX(/=) return *this; }
    PAI  OP+    (CTY x) const  { PAI y(*this); y+=x; return y; }
    PAI  OP-    (CTY x) const  { PAI y(*this); y-=x; return y; }
    PAI  OP*    (CTY x) const  { PAI y(*this); y*=x; return y; }
    PAI  OP/    (CTY x) const  { PAI y(*this); y/=x; return y; }
    PAI& OP=    (CTY*b)        { FT(a,b,=)    return *this; }
    PAI& OP+=   (CTY*b)        { FT(a,b,+=)   return *this; }
    PAI& OP-=   (CTY*b)        { FT(a,b,-=)   return *this; }
    PAI  OP+    (CTY*b) const  { PAI y(*this); y+=b; return y; }
    PAI  OP-    (CTY*b) const  { PAI y(*this); y-=b; return y; }
    PAI& OP=    (CPA&c)        { FT(a,c.a,=)  return *this; }
    PAI& OP+=   (CPA&c)        { FT(a,c.a,+=) return *this; }
    PAI& OP-=   (CPA&c)        { FT(a,c.a,-=) return *this; }
    PAI  OP+    (CPA&c) const  { PAI y(*this); y+=c; return y; }
    PAI  OP-    (CPA&c) const  { PAI y(*this); y-=c; return y; }
    PAI  OP-    ()      const  { PAI y(TY(0)); return y-= *this; }
    TY   OP^    (CPA&c)        { return a[0]*c[1] - a[1]*c[0]; }
         OP TY* ()             { return a; }
         OP CTY*()      const  { return a; }
    PAI& add_mul(CPA&c, CTY x) { a[0]+=x*c[0]; a[1]+=x*c[1]; return *this; }
    PAI& sub_mul(CPA&c, CTY x) { a[0]-=x*c[0]; a[1]-=x*c[1]; return *this; }
    PAI& negate ()             { FT(a,a,=-) return *this; }
    void apply  (TY(*Y)(CTY&)) { a[0]=Y(a[0]); a[1]=Y(a[1]); }
    void connect(CPA &c, TY(*Y)(CTY&,CTY&)) { a[0]=Y(a[0],c[0]); 
                                              a[1]=Y(a[1],c[1]); }
    TY   dist_sq(CPA &c) const { return square(a[0]-c[0]) + square(a[1]-c[1]); }
    TY   sum_sq (CPA &c) const { return square(a[0]+c[0]) + square(a[1]+c[1]); }
    TY   norm   () const       { return a[0]*a[0]+a[1]*a[1]; }
    TY   OP*    (CPA &c) const { return a[0]*c[0]+a[1]*c[1]; }
    TY   maxnorm() const       { register TY b=abs(a[0]),x=b;
    if(x<(b=abs(a[1]))) x=b; return x; }
    bool OP==   (CPA &c) const { if(a[0]!=c[0] || a[1]!=c[1]) return false;
    return true; }
    bool OP==   (CTY  x) const { if(a[0]!=x || a[1]!=x) return false;
                                 return true; }
    bool OP!=   (CPA &c) const { return !(*this==c); }
    bool OP!=   (CTY  x) const { return !(*this==x); }
  };

  //////////////////////////////////////////////////////////////////////////////
  // inline members of class paiR and more

#define TM  template<class TY> inline

  TM PAI<TY> OP+  (CTY x, CPA<TY>&y)      { return y+x; }
  TM PAI<TY> OP-  (CTY x, CPA<TY>&y)      { return -(y-x); }
  TM PAI<TY> OP*  (CTY x, CPA<TY>&y)      { return y*x; }
  TM TY  dist_sq  (CPA<TY>&x,CPA<TY>&y)   { return x.dist_sq(y); }
  TM TY  norm     (CPA<TY>&x)             { return x.norm(); }
  TM TY  maxnorm  (CPA<TY>&x)             { return x.maxnorm(); }
  TM TY  sum_sq   (CPA<TY>&x, CPA<TY>&y)  { return x.sum_sq(y); }
#ifndef C_AND_F
  TM std::ostream& OP<<(std::ostream& s, CPA<TY>&a) { return s<<a[0]<<' '<<a[1];}
  TM std::istream& OP>>(std::istream& s, PAI<TY>&a) {
    TY x[2];
    char c=0;    s>>c;
    if(c == '(') { s>>x[0]>>x[1]; s>>c; if(c!=')') s.clear(std::ios::badbit); }
    else         { s.putback(c); s>>x[0]>>x[1]; }
    if(s) a=x;
    return s;
  }
#endif // C_AND_F
}
////////////////////////////////////////////////////////////////////////////////
#undef  FTX
#undef  FT
#undef  CTY
#undef  PAI
#undef  CPA
#undef  CI
#undef  CUI
#undef  CUS
#undef  TM
#undef  OP
////////////////////////////////////////////////////////////////////////////////
#endif  // included_paiR_h
