// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// stat.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class TreeStats           fully inline                                      |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_stat_h
#define included_stat_h

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

////////////////////////////////////////////////////////////////////////////////
// class TreeStats
namespace fasttree {
  class TreeStats {
  private:
    typedef unsigned int uint;
    uint N_BB, N_CB, N_CC, N_MB;
  public:
    // 1 recording
    void reset    () { N_BB=0; N_CB=0; N_CC=0; N_MB=0; }
    void record_BB() { N_BB++; }
    void record_CB() { N_CB++; }
    void record_CC() { N_CC++; }
    void record_MB() { N_MB++; }
    // 2 reporting
    const uint& BB_interactions() const { return N_BB; }
    const uint& MB_interactions() const { return N_MB; }
    const uint& CB_interactions() const { return N_CB; }
    const uint& CC_interactions() const { return N_CC; }
    uint  total_interactions   () const { return N_BB+N_CB+N_CC+N_MB; }
    real BB_ia_part()const{ return 100*BB_interactions()/total_interactions();}
    real MB_ia_part()const{ return 100*MB_interactions()/total_interactions();}
    real CB_ia_part()const{ return 100*CB_interactions()/total_interactions();}
    real CC_ia_part()const{ return 100*CC_interactions()/total_interactions();}
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_stat_h
