// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tn2D.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class TY> class sym22         fully inline                         |
// template<class TY> class sym222        fully inline                         |
// some global operators                  fully inline                         |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_tn2D_h
#define included_tn2D_h

#ifndef included_pair_h
#include <tree/paiR.h>
#endif

////////////////////////////////////////////////////////////////////////////////

#define CTY         const TY
#define CI          const int
#define CUI         const unsigned int
#define CUS         const unsigned short
#define PAI         paiR
#define CPA         const paiR
#define OP          operator

namespace fasttree {

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::sym22 for a symmetric 2x2 matrix                           
  //////////////////////////////////////////////////////////////////////////////
#define SYM         sym22
#define CSY         const sym22
#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;

  template<class TY>
  class SYM {
  private:
    TY a[3];
    //   element mapping is as follows: 0 -> 0,0
    //                                  1 -> 0,1 & 1,0
    //                                  2 -> 1,1
  public:
    SYM          ()               {}
    SYM          (CTY x)          { FTX(=) }
    SYM          (CSY&c)          { FT (a,c.a,=) }
    SYM& unity   ()               { a[0]=a[2]=TY(1); a[1]=TY(0); return *this; }
         OP TY*  ()               { return a; }
         OP CTY* () const         { return a; }
    SYM& OP=     (CTY x)          { FTX(=)  return *this; }
    SYM& OP*=    (CTY x)          { FTX(*=) return *this; }
    SYM& OP/=    (CTY x)          { FTX(/=) return *this; }
    SYM  OP*     (CTY x)    const { SYM y(*this); y*=x; return y; }
    SYM  OP/     (CTY x)    const { SYM y(*this); y/=x; return y; }
    SYM& OP=     (CSY&c)          { FT (a,c.a,=)  return *this; }
    SYM& OP+=    (CSY&c)          { FT (a,c.a,+=) return *this; }
    SYM& OP-=    (CSY&c)          { FT (a,c.a,-=) return *this; }
    SYM  OP+     (CSY&c)    const { SYM y(*this); y+=c; return y; }
    SYM  OP-     (CSY&c)    const { SYM y(*this); y-=c; return y; }
    SYM& OP=     (CTY*b)          { FT(a,b,=)  return *this; }
    SYM& OP+=    (CTY*b)          { FT(a,b,+=) return *this; }
    SYM& OP-=    (CTY*b)          { FT(a,b,-=) return *this; }
    SYM  OP+     (CTY*b)    const { SYM y(*this); y+=b; return y; }
    SYM  OP-     (CTY*b)    const { SYM y(*this); y-=b; return y; }
    CTY& OP[]    (CI  i)    const { return a[i]; }
    CTY& OP[]    (CUI i)    const { return a[i]; }
    CTY& OP[]    (CUS i)    const { return a[i]; }
    TY&  OP[]    (CI  i)          { return a[i]; }
    TY&  OP[]    (CUI i)          { return a[i]; }
    TY&  OP[]    (CUS i)          { return a[i]; }
    CTY& OP()    (CI  i, CI  j) const { return a[i+j]; }
    CTY& OP()    (CUI i, CUI j) const { return a[i+j]; }
    CTY& OP()    (CUS i, CUS j) const { return a[i+j]; }
    TY&  OP()    (CI  i, CI  j)       { return a[i+j]; }
    TY&  OP()    (CUI i, CUI j)       { return a[i+j]; }
    TY&  OP()    (CUS i, CUS j)       { return a[i+j]; }
    void negate  ()               { FT(a,a,=-) }
    TY   trace   ()         const { return a[0] + a[2]; }
    bool OP==    (CSY&c)    const { if(a[0]!=c[0] || a[1]!=c[1] || a[2]!=c[2])
                                      return false;
                                    return true; }
    bool OP!=    (CSY&c)    const { return !(*this==c); }
    PAI<TY> OP*  (CPA<TY>& x) const
      // x_i * A_ij
    {
      register TY w[2] = { a[0]*x[0]+a[1]*x[1], a[1]*x[0]+a[2]*x[1] };
      return PAI<TY>(w);
    }
    TY   contract(CPA<TY>& x) const
      // Xi*Xj*Aij
    {
      return x[0] * (a[0]*x[0] + a[1]*x[1]) +
	     x[1] * (a[1]*x[0] + a[2]*x[1]);
    }
    TY   contract(CSY<TY>& M) const
      // Aij*Bij
    {
      register TY x =     a[1]*M[1];
      return      x + x + a[0]*M[0] + a[2]*M[2];
    }
    SYM& outprod (CPA<TY>& x)
      // Xi*Xj
    {
      a[0]=x[0]*x[0];
      a[1]=x[0]*x[1];
      a[2]=x[1]*x[1];
      return *this;
    }
    SYM& outprod (CPA<TY>& x, CPA<TY>& y)
      // Xi*Yj + Yi*Xj
    {
      a[0]=x[0]*y[0];
      a[0]+=a[0];
      a[1]=x[0]*y[1] + x[1]*y[0];
      a[2]=x[1]*y[1];
      a[2]+=a[2];
      return *this;
    }
    SYM& quadrup (CPA<TY>& x, CTY A, CTY B)
      // a*Dij + b*Xi*Xj
    {
      register TY Bx0 = B*x[0];
      a[0]=A+Bx0*x[0];
      a[1]=Bx0*x[1];
      a[2]=A+B*x[1]*x[1];
      return *this;
    }
  };

#undef  FTX
#undef  FT

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::sym222 for an symmetric 2D tensor with 3 indices           
  //////////////////////////////////////////////////////////////////////////////

#define SYT     sym222
#define CST     const sym222

#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2]; a[3] op b[3];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;    a[3] op x;

  template<class TY>
  class SYT {
  private:
    TY a[4];
    // element mapping is as follows: 0 -> 0,0,0
    //                                1 -> 0,0,1
    //                                2 -> 0,1,1
    //                                3 -> 1,1,1
  public:
    SYT             ()               {}
    SYT             (CTY x)          { FTX(=) }
    SYT             (CST&c)          { FT(a,c.a,=) }
    SYT& unity      ()               { a[0]=a[3]=TY(1); a[1]=a[2]=TY(0);
                                       return *this; }
            OP TY*  ()               { return a; }
            OP CTY* () const         { return a; }
    SYT&    OP=     (CTY x)          { FTX(=)  return *this; }
    SYT&    OP*=    (CTY x)          { FTX(*=) return *this; }
    SYT&    OP/=    (CTY x)          { FTX(/=) return *this; }
    SYT     OP*     (CTY x)    const { SYT y(*this); y*=x; return y; }
    SYT     OP/     (CTY x)    const { SYT y(*this); y/=x; return y; }
    SYT&    OP=     (CST&c)          { FT(a,c.a,=)  return *this; }
    SYT&    OP+=    (CST&c)          { FT(a,c.a,+=) return *this; }
    SYT&    OP-=    (CST&c)          { FT(a,c.a,-=) return *this; }
    SYT     OP+     (CST&c)    const { SYT y(*this); y+=c; return y; }
    SYT     OP-     (CST&c)    const { SYT y(*this); y-=c; return y; }
    SYT&    OP=     (CTY*b)          { FT(a,b,=)  return *this; }
    SYT&    OP+=    (CTY*b)          { FT(a,b,+=) return *this; }
    SYT&    OP-=    (CTY*b)          { FT(a,b,-=) return *this; }
    SYT     OP+     (CTY*b)    const { SYT y(*this); y+=b; return y; }
    SYT     OP-     (CTY*b)    const { SYT y(*this); y-=b; return y; }
    CTY&    OP[]    (CI  i)    const { return a[i]; }
    CTY&    OP[]    (CUI i)    const { return a[i]; }
    CTY&    OP[]    (CUS i)    const { return a[i]; }
    TY&     OP[]    (CI  i)          { return a[i]; }
    TY&     OP[]    (CUI i)          { return a[i]; }
    TY&     OP[]    (CUS i)          { return a[i]; }
    CTY&    OP()    (CI  i, CI  j, CI  k) const { return a[i+j+k]; }
    CTY&    OP()    (CUI i, CUI j, CUI k) const { return a[i+j+k]; }
    CTY&    OP()    (CUS i, CUS j, CUS k) const { return a[i+j+k]; }
    TY&     OP()    (CI  i, CI  j, CI  k)       { return a[i+j+k]; }
    TY&     OP()    (CUI i, CUI j, CUI k)       { return a[i+j+k]; }
    TY&     OP()    (CUS i, CUS j, CUS k)       { return a[i+j+k]; }
    void    negate  ()               { FT(a,a,=-) }
    bool    OP==    (CST&c)    const { if(a[0]!=c[0] || a[1]!=c[1] ||
					  a[2]!=c[2] || a[3]!=c[3] ) return false;
                                       return true; }
    bool    OP!=    (CST&c)    const { return !(*this==c); }
    PAI<TY> trace   ()         const
      // P_ijk*D_jk
    {
      register TY w[3] = { a[0]+a[2], a[1]+a[3]};
      return PAI<TY>(w);
    }
    TY      trace   (CPA<TY>& x) const
      // trace(x*A)
    {
      return   x[0]*(a[0]+a[2]) + x[1]*(a[1]+a[3]);
    }
    PAI<TY> OP*     (CSY<TY>& M) const
      // A_ijk*Mij
    {
      register PAI<TY> W;
      W[0] = a[1]*M[1];
      W[0] += W[0] + a[0]*M[0] + a[2]*M[2];
      W[1] = a[2]*M[1];
      W[1] += W[1] + a[1]*M[0] + a[3]*M[2];
      return W;
    }
    SYM<TY> OP*     (CPA<TY>& x) const
      // A_ijk*Xi
    {
      register SYM<TY> M;
      M[0] =a[0]*x[0] +a[1]*x[1];
      M[1] =a[1]*x[0] +a[2]*x[1];
      M[2] =a[2]*x[0] +a[3]*x[1];
      return M;
    }
    SYT&    outprod (CPA<TY>& x)
      // Xi*Xj*Xk
    {
      register TY tmp=x[0]*x[0];
      a[0]= tmp*x[0];
      a[1]= tmp*x[1];
      tmp =x[1]*x[1];
      a[2]= tmp*x[0];
      a[3]= tmp*x[1];
      return *this;
    }
    SYT&    outprod (CPA<TY>& x, CSY<TY>& M)
      // Xi*Mjk +Xj*Mik +Xk*Mij
    {
      a[0] =M[0]*x[0];
      a[0]+=a[0]+a[0];
      a[1] =M[1]*x[0];
      a[1]+=a[1]+M[0]*x[1];
      a[2] =M[1]*x[1];
      a[1]+=a[1]+M[2]*x[0];
      a[3] =M[2]*x[1];
      a[3]+=a[3]+a[3];
      return *this;
    } 
    SYT&    quadrup (CPA<TY>& x, CTY A)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi);
    {
      register TY ai=A*x[0];
      a[0]=ai+ai+ai;
      a[2]=ai;
      ai=A*x[1];
      a[3]=ai+ai+ai;
      a[1]=ai;
      return *this;
    }
    SYT&    quadrup (CPA<TY>& x, CTY A, CTY B)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi) + b Xi Xj Xk;
    {
      register TY A3=A+A+A, tmp=B*x[0]*x[0];
      a[0] =(A3+tmp)*x[0];
      a[1] =(A +tmp)*x[1];
      tmp  =B*x[1]*x[1];
      a[2] =(A +tmp)*x[0]; 
      a[3] =(A3+tmp)*x[1];
      return *this;
    }
  };

#undef  FTX
#undef  FT

  //////////////////////////////////////////////////////////////////////////////
  // some non-member operators                                                  
  //////////////////////////////////////////////////////////////////////////////

#define TM          template<class TY> inline

  TM PAI<TY> OP*  (CPA<TY>& X, CSY<TY>& M) { return M*X; }
  TM SYM<TY> OP*  (CPA<TY>& X, CST<TY>& N) { return N*X; }
  TM TY      trace(CSY<TY>& M)             { return M.trace(); }
  TM TY      trace(CST<TY>& M, CPA<TY>& x) { return M.trace(x); }
  TM PAI<TY> trace(CST<TY>& M)             { return M.trace(); }
#ifndef C_AND_F
  TM std::ostream& OP<< (std::ostream& s, CSY<TY>&M)
  { return s<<M[0]<<' '<<M[1]<<' '<<M[2]; }

  TM std::ostream& OP<< (std::ostream& s, CST<TY>&N) 
  {  return s<<N[0]<<' '<<N[1]<<' '<<N[2]<<' '<<N[3]; }
#endif // C_AND_F

}
////////////////////////////////////////////////////////////////////////////////
#undef  CTY
#undef  PAI
#undef  SYM
#undef  CPA
#undef  CSY
#undef  SYT
#undef  CST
#undef  CI
#undef  CUI
#undef  CUS
#undef  TM
#undef  OP
////////////////////////////////////////////////////////////////////////////////
#endif  // included_tn2D_h
