// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tn3D.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class TY> class sym33            fully inline                      |
// template<class TY> class sym333           fully inline                      |
// some global operators                     fully inline                      |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_tn3D_h
#define included_tn3D_h

#ifndef included_trip_h
#include <tree/trip.h>
#endif

////////////////////////////////////////////////////////////////////////////////

#define CTY         const TY
#define CI          const int
#define CUI         const unsigned int
#define CUS         const unsigned short
#define TRI         tripel
#define CTR         const tripel
#define OP          operator

namespace fasttree {

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::sym33 for a symmetric 3x3 matrix                           
  //////////////////////////////////////////////////////////////////////////////
#define SYM         sym33
#define CSY         const sym33
#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];			\
                    a[3] op b[3]; a[4] op b[4]; a[5] op b[5];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;			\
                    a[3] op x;    a[4] op x;    a[5] op x;
#define OPIJ        if(i>j) return (*this)(j,i);				\
                    switch(i+i+j) { 						\
                    case 0:  return a[0];					\
                    case 1:  return a[1];					\
                    case 2:  return a[2];					\
                    case 3:  return a[3];					\
                    case 4:  return a[4];					\
                    default: return a[5]; }

  template<class TY>
  class SYM {
  private:
    TY a[6];
//   element mapping is as follows: 0 -> 0,0
//                                  1 -> 0,1 & 1,0
//                                  2 -> 0,2 & 2,0
//                                  3 -> 1,1
//                                  4 -> 1,2 & 2,1
//                                  5 -> 2,2 & 2,2
  public:
    SYM          ()               {}
    SYM          (CTY x)          { FTX(=) }
    SYM          (CSY&c)          { FT (a,c.a,=) }
    SYM& unity   ()               { a[0]=a[3]=a[5]=TY(1); a[1]=a[2]=a[4]=TY(0);
                                    return *this; }
         OP TY*  ()               { return a; }
         OP CTY* () const         { return a; }
    SYM& OP=     (CTY x)          { FTX(=)  return *this; }
    SYM& OP*=    (CTY x)          { FTX(*=) return *this; }
    SYM& OP/=    (CTY x)          { FTX(/=) return *this; }
    SYM  OP*     (CTY x)    const { SYM y(*this); y*=x; return y; }
    SYM  OP/     (CTY x)    const { SYM y(*this); y/=x; return y; }
    SYM& OP=     (CSY&c)          { FT (a,c.a,=)  return *this; }
    SYM& OP+=    (CSY&c)          { FT (a,c.a,+=) return *this; }
    SYM& OP-=    (CSY&c)          { FT (a,c.a,-=) return *this; }
    SYM  OP+     (CSY&c)    const { SYM y(*this); y+=c; return y; }
    SYM  OP-     (CSY&c)    const { SYM y(*this); y-=c; return y; }
    SYM& OP=     (CTY*b)          { FT(a,b,=)  return *this; }
    SYM& OP+=    (CTY*b)          { FT(a,b,+=) return *this; }
    SYM& OP-=    (CTY*b)          { FT(a,b,-=) return *this; }
    SYM  OP+     (CTY*b)    const { SYM y(*this); y+=b; return y; }
    SYM  OP-     (CTY*b)    const { SYM y(*this); y-=b; return y; }
    CTY& OP[]    (CI  i)    const { return a[i]; }
    CTY& OP[]    (CUI i)    const { return a[i]; }
    CTY& OP[]    (CUS i)    const { return a[i]; }
    TY&  OP[]    (CI  i)          { return a[i]; }
    TY&  OP[]    (CUI i)          { return a[i]; }
    TY&  OP[]    (CUS i)          { return a[i]; }
    CTY& OP()    (CI  i, CI  j) const { OPIJ }
    CTY& OP()    (CUI i, CUI j) const { OPIJ }
    CTY& OP()    (CUS i, CUS j) const { OPIJ }
    TY&  OP()    (CI  i, CI  j)       { OPIJ }
    TY&  OP()    (CUI i, CUI j)       { OPIJ }
    TY&  OP()    (CUS i, CUS j)       { OPIJ }
    void negate  ()               { FT(a,a,=-) }
    TY   trace   ()         const { return a[0] + a[3] + a[5]; }
    bool OP==    (CSY&c)    const { if(a[0]!=c[0] || a[1]!=c[1] ||
				       a[2]!=c[2] || a[3]!=c[3] ||
				       a[4]!=c[4] || a[5]!=c[5] ) return false;
                                    return true; }
    bool OP!=    (CSY&c)    const { return !(*this==c); }
    TRI<TY> OP*  (CTR<TY>& x) const
      // x_i * A_ij
    {
      register TY w[3] = { a[0]*x[0] + a[1]*x[1] + a[2]*x[2],
			   a[1]*x[0] + a[3]*x[1] + a[4]*x[2], 
			   a[2]*x[0] + a[4]*x[1] + a[5]*x[2] }; 
      return TRI<TY>(w);
    }
    TY   contract(CTR<TY>& x) const
      // Xi*Xj*Aij
    {
      return x[0] * (a[0]*x[0] + a[1]*x[1] + a[2]*x[2]) +
   	     x[1] * (a[1]*x[0] + a[3]*x[1] + a[4]*x[2]) + 
     	     x[2] * (a[2]*x[0] + a[4]*x[1] + a[5]*x[2]);
    }
    TY   contract(CSY<TY>& M) const
      // Aij*Bij
    {
      register TY x =     a[1]*M[1] + a[2]*M[2] + a[4]*M[4];
      return      x + x + a[0]*M[0] + a[3]*M[3] + a[5]*M[5];
    }
    SYM& outprod (CTR<TY>& x)
      // Xi*Xj
    {
      a[0]=x[0]*x[0];
      a[1]=x[0]*x[1];
      a[2]=x[0]*x[2]; 
      a[3]=x[1]*x[1];
      a[4]=x[1]*x[2];
      a[5]=x[2]*x[2];
      return *this;
    }
    SYM& outprod (CTR<TY>& x, CTR<TY>& y)
      // Xi*Xj + Yi*Xj
    {
      a[0] =x[0]*y[0];
      a[0]+=a[0];
      a[1] =x[0]*y[1] + x[1]*y[0];
      a[2] =x[0]*y[2] + x[2]*y[0];
      a[3] =x[1]*y[1];
      a[3]+=a[3];
      a[4] =x[1]*y[2] + x[2]*y[1];
      a[5] =x[2]*y[2];
      a[5]+=a[5];
      return *this;
    }
    SYM& quadrup (CTR<TY>& x, CTY A, CTY B)
      // a*Dij + b*Xi*Xj
    {
      register TY tmp=B*x[0];
      a[0]= A+tmp*x[0];
      a[1]=   tmp*x[1];
      a[2]=   tmp*x[2];
      tmp = B*x[1];
      a[3]= A+tmp*x[1];
      a[4]=   tmp*x[2];
      a[5]= A+B*x[2]*x[2];
      return *this;
    }
  };

#undef  FTX
#undef  FT
#undef  OPIJ

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::sym333 for an symmetric 3D tensor with 3 indices           
  //////////////////////////////////////////////////////////////////////////////

#define SYT     sym333
#define CST     const sym333

#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];			\
                    a[3] op b[3]; a[4] op b[4]; a[5] op b[5];			\
                    a[6] op b[6]; a[7] op b[7]; a[8] op b[8]; a[9] op b[9];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;			\
                    a[3] op x;    a[4] op x;    a[5] op x;			\
                    a[6] op x;    a[7] op x;    a[8] op x;    a[9] op x;
#define OPIJK	    if(i>j)        { if(j>k) return (*this)(k,j,i);		\
                                     if(i>k) return (*this)(j,k,i);		\
                                        else return (*this)(j,i,k);		\
                    } else if(j>k) { if(i>k) return (*this)(k,i,j);		\
                                        else return (*this)(i,k,j);		\
                    } else { switch(i+i+i+i+j+j+k) {				\
                             case  0: return a[0];				\
                             case  1: return a[1];				\
                             case  2: return a[2];				\
                             case  3: return a[3];				\
                             case  4: return a[4];				\
                             case  6: return a[5];				\
                             case  7: return a[6];				\
                             case  8: return a[7];				\
                             case  9: return a[8];				\
                             default: return a[9]; } }

  template<class TY>
  class SYT {
  private:
    TY a[10];
// element mapping is as follows: 0 -> 0,0,0
//                                1 -> 0,0,1
//                                2 -> 0,0,2
//                                3 -> 0,1,1
//                                4 -> 0,1,2
//                                5 -> 0,2,2
//                                6 -> 1,1,1
//                                7 -> 1,1,2
//                                8 -> 1,2,2
//                                9 -> 2,2,2
  public:
    SYT             ()               {}
    SYT             (CTY x)          { FTX(=) }
    SYT             (CST&c)          { FT(a,c.a,=) }
    SYT& unity      ()               { a[1]=a[2]=a[3]=a[4]=a[5]=a[7]=a[8]=TY(0);
                                       a[0]=a[6]=a[9]=TY(1); return *this; }
            OP TY*  ()               { return a; }
            OP CTY* () const         { return a; }
    SYT&    OP=     (CTY x)          { FTX(=)  return *this; }
    SYT&    OP*=    (CTY x)          { FTX(*=) return *this; }
    SYT&    OP/=    (CTY x)          { FTX(/=) return *this; }
    SYT     OP*     (CTY x)    const { SYT y(*this); y*=x; return y; }
    SYT     OP/     (CTY x)    const { SYT y(*this); y/=x; return y; }
    SYT&    OP=     (CST&c)          { FT(a,c.a,=)  return *this; }
    SYT&    OP+=    (CST&c)          { FT(a,c.a,+=) return *this; }
    SYT&    OP-=    (CST&c)          { FT(a,c.a,-=) return *this; }
    SYT     OP+     (CST&c)    const { SYT y(*this); y+=c; return y; }
    SYT     OP-     (CST&c)    const { SYT y(*this); y-=c; return y; }
    SYT&    OP=     (CTY*b)          { FT(a,b,=)  return *this; }
    SYT&    OP+=    (CTY*b)          { FT(a,b,+=) return *this; }
    SYT&    OP-=    (CTY*b)          { FT(a,b,-=) return *this; }
    SYT     OP+     (CTY*b)    const { SYT y(*this); y+=b; return y; }
    SYT     OP-     (CTY*b)    const { SYT y(*this); y-=b; return y; }
    CTY&    OP[]    (CI  i)    const { return a[i]; }
    CTY&    OP[]    (CUI i)    const { return a[i]; }
    CTY&    OP[]    (CUS i)    const { return a[i]; }
    TY&     OP[]    (CI  i)          { return a[i]; }
    TY&     OP[]    (CUI i)          { return a[i]; }
    TY&     OP[]    (CUS i)          { return a[i]; }
    CTY&    OP()    (CI  i, CI  j, CI  k) const { OPIJK }
    CTY&    OP()    (CUI i, CUI j, CUI k) const { OPIJK }
    CTY&    OP()    (CUS i, CUS j, CUS k) const { OPIJK }
    TY&     OP()    (CI  i, CI  j, CI  k)       { OPIJK }
    TY&     OP()    (CUI i, CUI j, CUI k)       { OPIJK }
    TY&     OP()    (CUS i, CUS j, CUS k)       { OPIJK }
    void    negate  ()               { FT(a,a,=-) }
    bool    OP==    (CST&c)    const { if(a[0]!=c[0] || a[1]!=c[1] ||
					  a[2]!=c[2] || a[3]!=c[3] ||
					  a[4]!=c[4] || a[5]!=c[5] ||
					  a[6]!=c[6] || a[7]!=c[7] ||
					  a[8]!=c[8] || a[9]!=c[9] ) return false;
    				       return true; }
    bool    OP!=    (CST&c)    const { return !(*this==c); }
    TRI<TY> trace   ()         const
      // P_ijk*D_jk
    {
      register TY w[3] = { a[0] + a[3] + a[5],
			   a[1] + a[6] + a[8],
			   a[2] + a[7] + a[9] };
      return TRI<TY>(w);
    }
    TY      trace   (CTR<TY>& x) const
      // trace(x*A)
    {
      return   x[0]*(a[0]+a[3]+a[5])
	     + x[1]*(a[1]+a[6]+a[8])
	     + x[2]*(a[2]+a[7]+a[9]);
    }
    TRI<TY> OP*     (CSY<TY>& M) const
      // A_ijk*Mij
    {
      register TRI<TY> W;
      register TY x=    a[1]*M[1]+a[2]*M[2]+a[4]*M[4];
      W[0]         =x+x+a[0]*M[0]+a[3]*M[3]+a[5]*M[5];
      x            =    a[3]*M[1]+a[4]*M[2]+a[7]*M[4];
      W[1]         =x+x+a[1]*M[0]+a[6]*M[3]+a[8]*M[5];
      x            =    a[4]*M[1]+a[5]*M[2]+a[8]*M[4];
      W[2]         =x+x+a[2]*M[0]+a[7]*M[3]+a[9]*M[5];
      return W;
    }
      SYM<TY> OP*     (CTR<TY>& x) const
	// A_ijk*Xi
    {
      register SYM<TY> M;
      M[0] =a[0]*x[0] +a[1]*x[1] +a[2]*x[2];
      M[1] =a[1]*x[0] +a[3]*x[1] +a[4]*x[2];
      M[2] =a[2]*x[0] +a[4]*x[1] +a[5]*x[2];
      M[3] =a[3]*x[0] +a[6]*x[1] +a[7]*x[2];
      M[4] =a[4]*x[0] +a[7]*x[1] +a[8]*x[2];
      M[5] =a[5]*x[0] +a[8]*x[1] +a[9]*x[2];
      return M;
    }
    SYT&    outprod (CTR<TY>& x)
      // Xi*Xj*Xk
    {
      register TY x00=x[0]*x[0], x11=x[1]*x[1], x22=x[2]*x[2];
      a[0]= x00*x[0];
      a[1]= x00*x[1];
      a[2]= x00*x[2];
      a[3]= x11*x[0];
      a[4]= x[0]*x[1]*x[2];
      a[5]= x22*x[0];
      a[6]= x11*x[1];
      a[7]= x11*x[2];
      a[8]= x22*x[1];
      a[9]= x22*x[2];
      return *this;
    }
    SYT&    outprod (CTR<TY>& x, CSY<TY>& M)
      // Xi*Mjk +Xj*Mik +Xk*Mij
    {
      a[0] =3*M[0]*x[0];
      a[6] =3*M[3]*x[1];
      a[9] =3*M[5]*x[2];
      a[1] =M[0]*x[1] +2*M[1]*x[0];
      a[2] =M[0]*x[2] +2*M[2]*x[0];
      a[3] =M[3]*x[0] +2*M[1]*x[1];
      a[5] =M[5]*x[0] +2*M[2]*x[2];
      a[7] =M[3]*x[2] +2*M[4]*x[1];
      a[8] =M[5]*x[1] +2*M[4]*x[2];
      a[4] =M[1]*x[2] +M[2]*x[1] +M[4]*x[0];
      return *this;
    } 
    SYT&    quadrup (CTR<TY>& x, CTY A)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi);
    { 
      register TY ai=A*x[0];
      a[0]=ai+ai+ai;
      a[3]=ai;
      a[5]=ai;
      ai  =A*x[1];
      a[6]=ai+ai+ai;
      a[1]=ai;
      a[8]=ai;
      ai  =A*x[2];
      a[9]=ai+ai+ai;
      a[2]=ai;
      a[7]=ai;
      a[4]=zero;
      return *this;
    }
    SYT&    quadrup (CTR<TY>& x, CTY A, CTY B)
      // Pijk = a (Dij Xk + Dik Xj + Djk Xi) + b Xi Xj Xk;
    {
      register TY A3=A+A+A, tmp=B*x[0]*x[0];
      a[0] = (A3+tmp)*x[0];
      a[1] = (A +tmp)*x[1];
      a[2] = (A +tmp)*x[2];
      tmp  = B*x[1]*x[1];
      a[3] = (A +tmp)*x[0];
      a[6] = (A3+tmp)*x[1];
      a[7] = (A +tmp)*x[2];
      tmp  = B*x[2]*x[2];
      a[5] = (A +tmp)*x[0];
      a[8] = (A +tmp)*x[1];
      a[9] = (A3+tmp)*x[2];
      a[4] = B*x[0]*x[1]*x[2];
      return *this;
    }
  };

#undef  FTX
#undef  FT
#undef  OPIJK

  //////////////////////////////////////////////////////////////////////////////
  // some non-member operators                                                  
  //////////////////////////////////////////////////////////////////////////////

#define TM          template<class TY> inline

  TM TRI<TY> OP*  (CTR<TY>& X, CSY<TY>& M) { return M*X; }
  TM SYM<TY> OP*  (CTR<TY>& X, CST<TY>& N) { return N*X; }
  TM TY      trace(CSY<TY>& M)             { return M.trace(); }
  TM TY      trace(CST<TY>& M, CTR<TY>& x) { return M.trace(x); }
  TM TRI<TY> trace(CST<TY>& M)             { return M.trace(); }
#ifndef C_AND_F
  TM std::ostream& OP<< (std::ostream& s, CSY<TY>&M)
  {
    return s<<M[0]<<' '<<M[1]<<' '<<M[2]<<' '<<M[3]<<' '<<M[4]<<' '<<M[5];
  }
  TM std::ostream& OP<< (std::ostream& s, CST<TY>&N) 
  {
    return s<<N[0]<<' '<<N[1]<<' '<<N[2]<<' '<<N[3]<<' '<<N[4]<<' '
	    <<N[5]<<' '<<N[6]<<' '<<N[7]<<' '<<N[8]<<' '<<N[9];
  }
#endif

}
////////////////////////////////////////////////////////////////////////////////
#undef  CTY
#undef  TRI
#undef  SYM
#undef  CTR
#undef  CSY
#undef  SYT
#undef  CST
#undef  CI
#undef  CUI
#undef  CUS
#undef  TM
#undef  OP
////////////////////////////////////////////////////////////////////////////////
#endif	// included_tn3D_h
