// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// trip.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1999-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class TY> class tripel            fully inline                     |
// some global operators                      fully inline                     |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_trip_h
#define included_trip_h

#ifndef included_inline_h
#include <tree/inline.h>
#endif

extern "C" double sqrt(double);

////////////////////////////////////////////////////////////////////////////////
// class tripel for an tripel of elements of type TY

#define CTY         const TY
#define CI          const int
#define CUI         const unsigned int
#define CUS         const unsigned short
#define TRI         tripel
#define CTR         const tripel
#define OP          operator
#define FR          friend
#define FT(a,b,op)  a[0] op b[0]; a[1] op b[1]; a[2] op b[2];
#define FTX(op)     a[0] op x;    a[1] op x;    a[2] op x;

namespace fasttree {
  template<class TY>
  class TRI {
  private:
    TY a[3];
  public:
    TY&  OP[]   (CI  i)       { return a[i]; }
    CTY& OP[]   (CI  i) const { return a[i]; }
    TY&  OP[]   (CUI i)       { return a[i]; }
    CTY& OP[]   (CUI i) const { return a[i]; }
    TY&  OP[]   (CUS i)       { return a[i]; }
    CTY& OP[]   (CUS i) const { return a[i]; }
    TRI         ()            {}
   ~TRI         ()            {}
    TRI         (CTY x)       { FTX(=) }
    TRI         (CTY*b)       { FT (a,b,=) }
    TRI         (CTR&c)       { FT (a,c.a,=) }
    TRI& OP=    (CTY x)       { FTX(=)  return *this; }
    TRI& OP*=   (CTY x)       { FTX(*=) return *this; }
    TRI& OP/=   (CTY x)       { FTX(/=) return *this; }
    TRI  OP+    (CTY x) const { TRI y(*this); y+=x; return y; }
    TRI  OP-    (CTY x) const { TRI y(*this); y-=x; return y; }
    TRI  OP*    (CTY x) const { TRI y(*this); y*=x; return y; }
    TRI  OP/    (CTY x) const { TRI y(*this); y/=x; return y; }
    TRI& OP=    (CTY*b)       { FT(a,b,=)    return *this; }
    TRI& OP+=   (CTY*b)       { FT(a,b,+=)   return *this; }
    TRI& OP-=   (CTY*b)       { FT(a,b,-=)   return *this; }
    TRI  OP+    (CTY*b) const { TRI y(*this); y+=b; return y; }
    TRI  OP-    (CTY*b) const { TRI y(*this); y-=b; return y; }
    TRI& OP=    (CTR&c)       { FT(a,c.a,=)  return *this; }
    TRI& OP+=   (CTR&c)       { FT(a,c.a,+=) return *this; }
    TRI& OP-=   (CTR&c)       { FT(a,c.a,-=) return *this; }
    TRI  OP+    (CTR&c) const { TRI y(*this); y+=c; return y; }
    TRI  OP-    (CTR&c) const { TRI y(*this); y-=c; return y; }
    TRI  OP^    (CTR&c) const { TRI z;
                                z[0] = a[1]*c.a[2] - a[2]*c.a[1];
				z[1] = a[2]*c.a[0] - a[0]*c.a[2];
				z[2] = a[0]*c.a[1] - a[1]*c.a[0];
				return z; }
    TRI  OP-    ()      const { TRI y(TY(0)); return y-= *this; }
         OP TY* ()            { return a; }
         OP CTY*()      const { return a; }
    TRI& add_mul(CTR&c, CTY x){ a[0]+=x*c[0]; a[1]+=x*c[1]; a[2]+=x*c[2];
                                return *this; }
    TRI& sub_mul(CTR&c, CTY x){ a[0]-=x*c[0]; a[1]-=x*c[1]; a[2]-=x*c[2];
                                return *this; }
    TRI& negate ()            { FT(a,a,=-) return *this; }
    void apply  (TY(*Y)(CTY&)){ a[0]=Y(a[0]); a[1]=Y(a[1]); a[2]=Y(a[2]); }
    void connect(CTR &c, TY(*Y)(CTY&,CTY&)) { a[0]=Y(a[0],c[0]); 
    a[1]=Y(a[1],c[1]); a[2]=Y(a[2],c[2]); }
    TY   norm   () const      { return a[0]*a[0]+a[1]*a[1]+a[2]*a[2]; }
    TY   dist_sq(CTR&c) const { return   square(a[0]-c[0])
				       + square(a[1]-c[1])
				       + square(a[2]-c[2]); }
    TY   sum_sq (CTR&c) const { return   square(a[0]+c[0])
				       + square(a[1]+c[1])
				       + square(a[2]+c[2]); }
    TY   OP*    (CTR&c) const { return a[0]*c[0]+a[1]*c[1]+a[2]*c[2]; }
    TY   maxnorm() const      { register TY b=abs(a[0]),x=b;
                                    if(x<(b=abs(a[1]))) x=b;
                                    if(x<(b=abs(a[2]))) x=b; return x; }
    bool OP==   (CTR&c) const { if(a[0]!=c[0] || a[1]!=c[1] || a[2]!=c[2])
      return false; return true; }
    bool OP==   (CTY x) const { if(a[0]!=x || a[1]!=x || a[2]!=x)
      return false; return true; }
    bool OP!=   (CTR&c) const { return !(*this==c); }
    bool OP!=   (CTY x) const { return !(*this==x); }
  };

  //////////////////////////////////////////////////////////////////////////////
  // some non-member functions

#define TM  template<class TY> inline

  TM TRI<TY> OP+  (CTY x, CTR<TY>&y)      { return y+x; }
  TM TRI<TY> OP-  (CTY x, CTR<TY>&y)      { return -(y-x); }
  TM TRI<TY> OP*  (CTY x, CTR<TY>&y)      { return y*x; }
  TM TY  norm     (CTR<TY>&x)             { return x.norm(); }
  TM TY  dist_sq  (CTR<TY>&x, CTR<TY>&y)  { return x.dist_sq(y); }
  TM TY  sum_sq   (CTR<TY>&x, CTR<TY>&y)  { return x.sum_sq(y); }
  TM TY  maxnorm  (CTR<TY>&x)             { return x.maxnorm(); }
#ifndef C_AND_F
  TM std::ostream& OP<<(std::ostream& s, CTR<TY>&a)
  {
    return s<<a[0]<<' '<<a[1]<<' '<<a[2];
  }
  TM std::istream& OP>>(std::istream& s, TRI<TY>&a) {
    TY x[3];
    char c=0;    s>>c;
    if(c == '(') { s>>x[0]>>x[1]>>x[2]; s>>c; if(c!=')')
      s.clear(std::ios::badbit); }
    else         { s.putback(c); s>>x[0]>>x[1]>>x[2]; }
    if(s) a=x;
    return s;
  }
#endif // C_AND_F
}
////////////////////////////////////////////////////////////////////////////////
#undef  FTX
#undef  FT
#undef  CTY
#undef  TRI
#undef  CTR
#undef  CI
#undef  CUI
#undef  CUS
#undef  TM
#undef  OP
////////////////////////////////////////////////////////////////////////////////
#endif	// included_trip_h
