//-----------------------------------------------------------------------------+
//                                                                             |
// YancNemo.cc                                                                 |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// written by Walter Dehnen, 2001                                              |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//           YancNemo -- Yet Another N-body Code, for                          |
//                       Not Everybody Must Observe                            |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0    28/05/2001   WD created with the help of NEMO caption PJT          |
// v 1.0.1  01/06/2001   WD bug in this file removed, added history output     |
// v 1.0.2  13/06/2001   WD Ncrit option added (before Ncrit==1)               |
// v 1.0.3  20/06/2001   WD tree::make() back to adding-leafs algorithm        |
//                          Coordinatesystem given with snapshot               |
// v 1.0.4  25/06/2001   WD bug (builder::link_tree()) removed (thanks to PJT) |
// v 1.0.5  28/06/2001   WD changes in tree::grow. Avoid box overflow.         |
//                          bug (in builder removed)                           |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef ALLOW_NEMO
#error "You need \"ALLOW_NEMO\" to compile YancNemo"
#endif

#include <yanc.h>
extern "C" {
#include <nemo.h>
}
#include <iostream.h>

using fasttree::yanc;

//------------------------------------------------------------------------------
string defv[] = {
  "in=???\n            input file                                ",
  "out=???\n           output file                               ",
  "tstop=10\n          integration time                          ",
  "step=1\n            output integration step                   ",
  "theta=0.5\n         opening parameter at M=M_root             ",
  "Ncrit=6\n           max # bodies in un-splitted cells         ",
  "eps=0.05\n          softening length                          ",
  "kernel=21\n         softening kernel                          ",
  "hmin=6\n            tau_min = (1/2)^hmin                      ",
  "Nlev=1\n            # time-step levels                        ",
  "Nref=16\n           # bodies in cell for density estimate     ",
  "fac=\n              factor in tau < fac / rho                 ",
  "give_pot=0\n        0/1:  output potentials                   ",
  "give_rho=0\n        0/1/2 output none/density/surface density ",
  "VERSION=1.0.5\n     28/June/2001 WD                           ",
  NULL};
string usage = "YancNemo -- Yet Another N-body Code, for\n"
               "            Not Everybody Must Observe";
//------------------------------------------------------------------------------
void nemo_main()
{
  string          in, out;
  register int    kern,hmin,givepot,giverho,Ns,Nr,Nc;
  register double theta,eps,fac=0.001,dTou,Tout,Tend;
  in     = getparam ("in");
  out    = getparam ("out");
  Tend   = getdparam("tstop");
  dTou   = getdparam("step");
  theta  = getdparam("theta");
  Nc     = getiparam("Ncrit");
  eps    = getdparam("eps");
  kern   = getiparam("kernel");
  hmin   = getiparam("hmin");
  Ns     = getiparam("Nlev");
  Nr     = getiparam("Nref");
  if(Ns > 1) {
    if(!hasvalue("fac") ) {
      cerr << "YancNemo: fac required if Nlev > 1\n";
      exit(1);
    }
    fac    = getdparam("fac");
  }
  givepot = getiparam("give_pot");
  giverho = getiparam("give_rho");
  yanc YANC(in,true,theta,Nc,eps,kern,hmin,givepot,giverho,Ns,Nr,fac);
  if(!YANC.okay()) exit(1);
  if(Tend <  YANC.initial_time()) {
    cerr << "YancNemo: DONE: tstop < t_ini\n";
    exit(1);
  }
  Tout = YANC.initial_time() + 0.999999*dTou;
  YANC.describe_nemo(cout,*(ask_history()));
  YANC.short_head   (cout);
  YANC.short_stats  (cout);
  YANC.open_nemo    (out);
  YANC.write_nemo   ();
  while(YANC.time() < Tend) {
    YANC.block_step();
    YANC.short_stats(cout);
    if(YANC.time() >= Tout) {
      YANC.write_nemo();
      Tout += dTou;
    }
  }
}
//---------------------end-of-YancNemo.cc------that's-it-!-----------------------
