//-----------------------------------------------------------------------------+
//                                                                             |
// buil.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <tree/buil.h>                             // class gardener            
#include <tree/cell.h>                             // class cell                
#include <tree/bman.h>                             // class manager<BOX>        
#include <tree/soul.h>                             // class soul                
#include <body.h>                                  // class body                
////////////////////////////////////////////////////////////////////////////////
extern "C" double floor(double);                   // forward declaration       
#define LoopDims for(register indx d=0; d<NDIM; d++)
#define LoopOcts for(register indx b=0; b<NSUB; b++)

#ifdef SUBCELL_LLIST
  -- BUG -- for this code to work, you must not define SUBCELL_LLIST in cell.h
#endif
#ifdef SUBSOULL_LLIST
  -- BUG -- for this code to work, you must not define SUBSOUL_LLIST in soul.h
#endif
////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::leaf                                                       
  //////////////////////////////////////////////////////////////////////////////
  struct leaf : public node {
    leaf *NEXT;
#ifdef USE_BODIES
    body *MYBODY;
    void  set_up  (body*B)
    {
      pos()=B->ppos();
      MYBODY=B;
    }
    void  set_soul(soul*S)
    {
      MYBODY->set_soul(S);
      S->link_and_set_up(MYBODY);
    }
#else  // USE_BODIES
    uint  MYELEM;
    void  set_up  (const areal*x[NDIM], const uint i)
    {
      MYELEM = i;
      LoopDims pos(d) = x[d][i]; 
    }
    void  set_soul(soul*S, const areal*m)
    {
      S->link_and_set_up(m[MYELEM],pos(),MYELEM);
    }
#endif // USE_BODIES
    leaf() {}
    friend const vect& cofm(const leaf* const& L) { return L->pos(); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::box                                                        
  //////////////////////////////////////////////////////////////////////////////
  const indx BIT[8] = {1,2,4,8,16,32,64,128};
  //////////////////////////////////////////////////////////////////////////////
  struct box : public node {
    real     RADIUS;                               // half size of the cube     
    indx     TYPE;                                 // bitfield: 1=cell, 0=leaf  
    bool     IS_ZOMBIE;                            // are we a zombie cell?     
    node    *OCT[NSUB];                            // octants                   
    uint     NUMBER;                               // number of leafs           
    leaf    *LEAFS;                                // linked list of leafs      
    //--------------------------------------------------------------------------
    bool marked_as_box  (const indx i) const {return TYPE & BIT[i]; }
    bool marked_as_leaf (const indx i) const {return !marked_as_box(i); }
    bool octant_is_empty(const indx i) const {return OCT[i]==0; }
    bool octant_is_box  (const indx i) const {return OCT[i] && marked_as_box(i);}
    bool octant_is_leaf (const indx i) const {return OCT[i] && marked_as_leaf(i);}
    void mark_as_box    (const indx i)       {TYPE |=  BIT[i]; }
    void mark_as_leaf   (const indx i)       {TYPE &= ~BIT[i]; }
    void reset_octants  ()                   { LoopOcts OCT[b]=0;}
    vect const&center        ()             const {return pos(); }
    vect&      center        ()                   {return pos(); }
    real const&center        (const indx i) const {return pos()[i]; }
    real&      center        (const indx i)       {return pos()[i]; }
    friend vect      &center (      box*B)        {return B->center();  }
    friend vect const&center (const box*B)        {return B->center();  }
    //--------------------------------------------------------------------------
    box() : TYPE(0), IS_ZOMBIE(true), NUMBER(0), LEAFS(0) {}
    //--------------------------------------------------------------------------
    indx octant          (const leaf* L) const     // I: leaf (must be in box)  
      // This routine returns the octant of leaf L.                             
    {
      const    indx bit[3]={1,2,4};                // constant bits             
      register indx oct=0;                         // return variable           
      LoopDims                                     // loop dimensions >         
	if(cofm(L)[d] > center(d)) oct |= bit[d];  //   xoring in bits <        
      return oct;                                  // return octant             
    }
    //--------------------------------------------------------------------------
    void addleaf_to_list(leaf *L)
      // This routines simply adds a leaf to the linked list of leafs.          
    {
      L->NEXT   = LEAFS;                           // add to linked list        
      LEAFS     = L;                               // update head of list       
      NUMBER ++;                                   // increment number          
    }
    //--------------------------------------------------------------------------
    void addleaf_to_octs(leaf *L)
      // This routines simply adds a leaf to the linked list of leafs.          
    {
      register indx b = octant(L);                 // find appropriate octant   
      OCT[b] = L;                                  // fill into octant          
      mark_as_leaf(b);                             // mark octant as leaf       
      NUMBER ++;                                   // increment number          
    }
    //--------------------------------------------------------------------------
    void shrink_to_octant(const indx i)            // I: octant                 
      // This routines changes our center & radius to that of the ith octant.   
    {
      const indx bit[3]={1,2,4};                   // const bits                
      RADIUS *= 0.5;                               // reduce radius             
      LoopDims                                     // loop dimensions     >     
	if(i & bit[d]) center(d) += RADIUS;        //   correct center up       
	else           center(d) -= RADIUS;        //   or down           <     
    }
    //--------------------------------------------------------------------------
    indx  N_subboxes()
      // This routine counts the number of sub-boxes.                           
    {
      register indx n=0;                           // index, counter            
      LoopOcts if(TYPE & BIT[b]) n++;              // loop sub-nodes, count     
      return n;                                    // return coutner            
    }
  };
  //----------------------------------------------------------------------------
  inline vect RootCenter(const vect& xmin, const vect& xmax)
  {
    // This routines returns, in each dimension, the nearest integer to the     
    // center between xmax and xmin.                                            
    register vect c = zero;                        // reset return position     
    LoopDims c[d]=int(half*(xmax[d]+xmin[d])+half);// find center position      
    return c;                                      // and return it             
  }
  //----------------------------------------------------------------------------
  inline real RootRadius(const vect& X, const vect& xmin, const vect& xmax)
  {
    // This routines returns the half-size R of the smallest cube, centered on X
    // that contains the points xmin and xmax.                                  
    register real R,D=zero;                        // R, distance to leaf       
    LoopDims {                                     // loop dimensions       >   
      R = max(abs(xmax[d]-X[d]),abs(xmin[d]-X[d]));//   distance to xmin, xmax  
      update_max(D,R);                             //   update maximum distance 
    }                                              // <                         
    return pow_int(two,int(one+LOG(D)/LOG(two)));  // return R                  
  }
  //----------------------------------------------------------------------------
}
////////////////////////////////////////////////////////////////////////////////
using namespace fasttree;                          // use methods of WD's tree  
////////////////////////////////////////////////////////////////////////////////
// class fasttree::builder                                                      
////////////////////////////////////////////////////////////////////////////////
const fasttree::uint& builder::N_allocated() const { return BM->N_allocated(); }
//------------------------------------------------------------------------------
const fasttree::uint& builder::N_used()      const { return BM->N_used(); }
//------------------------------------------------------------------------------
#ifdef USE_BODIES
builder::builder(const uint nc, body* b0, const uint nb)
  : NCRIT(nc), NZOMBIES(0)
{
  const    body* bn=b0+nb;                         // beyond last body          
  register body* bi=b0;                            // current body              
  L0 = new leaf[nb];                               // allocate leafs            
  register leaf *Li=L0;                            // current leaf              
  XMAX = XMIN = tpos(b0);                          // reset x_min & x_max       
  for(; bi<bn; bi++) if(is_in_tree(bi)) {          // loop active bodies    >   
    Li->set_up(bi);                                //   initialize leaf         
    LoopDims {                                     //   loop dimensions     >   
      update_max(XMAX[d], cofm(Li)[d]);            //     update x_max          
      update_min(XMIN[d], cofm(Li)[d]);            //     update x_min          
    }                                              //   <                       
    Li++;                                          //   increment current leaf  
  }                                                // <                         
  LN     = Li;                                     // set: beyond last leaf     
  NLEAFS = int(LN-L0);                             // set: # active leafs       
  BM     = new manager<box>(NLEAFS);               // initialize box manager    
}
//------------------------------------------------------------------------------
#else  // USE_BODIES
builder::builder(const uint nc, const areal* x[NDIM], const areal *m,
		 const int* f, const uint nb)
  : NCRIT(nc), NZOMBIES(0), MASSES(m)
{
  register uint i;                                 // index                     
  L0 = new leaf[nb];                               // allocate leafs            
  register leaf *Li=L0;                            // current          leaf     
  LoopDims XMAX[d] = XMIN[d] = x[d][0];            //   reset x_min & x_max <   
  for(i=0; i<nb; i++) if(is_in_tree(f[i])) {       // loop active bodies    >   
    Li->set_up(x,i);                               //   initialize leaf         
    LoopDims {                                     //   loop dimensions     >   
      update_max(XMAX[d], cofm(Li)[d]);            //     update x_max          
      update_max(XMIN[d], cofm(Li)[d]);            //     update x_min          
    }                                              //   <                       
    Li++;                                          //   increment current leaf  
  }                                                // <                         
  LN     = Li;                                     // set: beyond last leaf     
  NLEAFS = int(LN-L0);                             // set: # active leafs       
  BM     = new manager<box>(NLEAFS);               // initialize box manager    
}
#endif // USE_BODIES
//==============================================================================
builder::~builder()                                // destructor                
{
  delete[] L0;                                     // de-allocate leafs         
  delete   BM;                                     // de-allocate boxes         
}
//------------------------------------------------------------------------------
bool builder::is_twig(const box* B) const
{
  return B->NUMBER <= NCRIT;                       // twig if not N > Ncrit     
}
//------------------------------------------------------------------------------
inline box* builder::new_box(const uint nl)
{
  return BM->new_box(nl);                          // get new box from manager 
}
//------------------------------------------------------------------------------
inline box* builder::make_subbox(const box* B, const indx i, const uint nl)
  // This routines provides a new box, whose center and radius are that of the  
  // ith octant of box B.                                                       
{
  register box* P = new_box(nl);                   // get box off the stock     
  P->RADIUS   = B->RADIUS;                         // copy radius of parent     
  P->center() = B->center();                       // copy center of parent     
  P->shrink_to_octant(i);                          // shrink to correct octant  
  return P;                                        // return new box            
}
//------------------------------------------------------------------------------
inline box* builder::make_subbox(const box* B, const indx i, leaf* L,
				 const uint nl)
  // This routines provides a new box, whose center and radius are that of the  
  // ith octant of box B and which contains leaf L as first entry.              
{
  register box *P = make_subbox(B,i,nl);           // make new sub-box          
  if(NCRIT>1) P->addleaf_to_list(L);               // add old leaf to list      
  else        P->addleaf_to_octs(L);               // or to appropriate oc      
  return P;                                        // return new box            
}
//------------------------------------------------------------------------------
inline void builder::split_box(      box  *P,      // I: box to be splitted     
			       const indx l,       // I: level                  
			       const uint nl)      // I: # leafs added so far   
  // This routines splits a box:                                                
  // The leafs in the linked list are sorted into octants. Octants with one     
  // leaf will just hold that leaf, octants with many leaf will be boxes with   
  // the leaf in the linked list.                                               
  // If all leafs happen to be in just one octant, the process is repeated      
  // recursively on the box of this octant.                                     
{
  if(l>DMAX) error("exceeding max depth");         // IF (level>DMAX) ERROR     
  register uint  NUM[NSUB];                        // array with number of leafs
  register indx  b,ne;                             // octant & counters         
  register box  *sub=0;                            // current sub-box           
  register leaf *Si,*Sn;                           // current & next leaf       
  do {                                             // DO until no zombie >      
    for(b=0; b<NSUB; b++) {                        //   loop octants  <         
      P->OCT[b] = 0;                               //     reset pointer         
      NUM[b]    = 0;                               //     reset counter         
    }                                              //   <                       
    Sn = P->LEAFS;                                 //   head of linked list     
    do {                                           //   loop linked list >      
      Si = Sn;                                     //     current leaf          
      Sn = Sn->NEXT;                               //     next leaf in list     
      b  = P->octant(Si);                          //     octant of current leaf
      Si->NEXT = static_cast<leaf*>(P->OCT[b]);    //     add leaf to list      
      P->OCT[b]= Si;                               //     update head of list   
      NUM[b] ++;                                   //     increase number       
    } while(Sn);                                   //   <                       
    P->LEAFS = 0;                                  //   reset list of sub-leafs 
    for(ne=b=0; b<NSUB; b++) if(NUM[b]) {          //   loop non-empty octants >
      ne++;                                        //     count them            
      if(NUM[b]>1) {                               //     IF(many leafs) >      
	sub = make_subbox(P,b,nl);                 //       make sub-box        
	sub->LEAFS = static_cast<leaf*>(P->OCT[b]);//       assign sub-box's    
	sub->NUMBER= NUM[b];                       //       leaf list & number  
	P->OCT[b]= sub;                            //       set octant=sub-box  
	P->mark_as_box(b);                         //       mark octant as box  
      }                                            //     <                     
    }                                              //   <                       
    if(ne!=1) P->IS_ZOMBIE = false;                //   mark P as non-zombie    
    P = sub;                                       //   set current box=sub-box 
  } while(ne==1);                                  // < while only 1 octant     
}
//------------------------------------------------------------------------------
void builder::addleafs(box *base)                  // add to this               
{
  register indx   b,l;                             // octant, level             
  register uint   nl=0;                            // counter                   
  register leaf  *Li,*Lo;                          // actual leaf loaded        
  register box   *P;                               // actual box                
  register node **oc;                              // actual sub-node           
  for(Li=L0; Li<LN; Li++,nl++) {                   // loop over all leafs >     
    P = base;                                      //   current box = base      
    for(l=0; l<DMAX; l++) {                        //   loop over boxes         
      if(NCRIT>1 && is_twig(P)) {                  //   IF(box == twig) >       
	P->addleaf_to_list(Li);                    //     add leaf to list      
	if(P->NUMBER > NCRIT) split_box(P,l,nl);   //     IF(N > NCRIT) split   
	break;                                     //     DONE with this leaf   
      } else {                                     //   < ELSE (box==branch) >  
	b  = P->octant(Li);                        //     leaf's octant         
	oc = P->OCT+b;                             //     pointer of octant     
	P->NUMBER++;                               //     increment number      
	if((*oc)==0) {                             //     IF octant empty >     
	  *oc = Li;                                //       assign leaf to it   
	  P->mark_as_leaf(b);                      //       mark octant as leaf 
	  P->IS_ZOMBIE = false;                    //       box cannot be zombie
	  break;                                   //       DONE with this leaf 
	} else if(P->marked_as_leaf(b)) {          //     < ELSE (octant=leaf) >
	  Lo = static_cast<leaf*>(*oc);            //       get old leaf        
	  P->mark_as_box(b);                       //       mark octant as box  
	  P = make_subbox(P,b,Lo,nl);              //       create sub-box      
	  *oc = P;                                 //       assign sub-box to oc
	} else                                     //     < ELSE (octant=box) > 
	  P  = static_cast<box*>(*oc);             //       set current box     
      }                                            //     <                     
    }                                              //   <                       
    if(l>=DMAX) error("exceeding max depth");      //   IF (level>DMAX) ERROR   
  }                                                // <                         
}
//------------------------------------------------------------------------------
void builder::build_tree(const vect*pc,            // I: center for root        
			 const int  dmax)          // I: max allowed tree depth 
  // This routine simply calls the above with the root box as argument.         
{
  DMAX  = dmax;                                    // max level allowed         
  register box *P= new_box(1);                     // get a new box: root       
  P->center() = pc? *pc:RootCenter(XMIN,XMAX);     // set root: center          
  P->RADIUS   = RootRadius(P->center(),XMIN,XMAX); // set root: radius          
  addleafs(P);                                     // call above, box=root      
  register uint nzom=0u;                           // zombie counter            
  for(register manager<box>::iterator              // loop boxes                
	I=BM->begin(); I!=BM->end(); ++I)          //   (using iterator)   >    
    if(!is_twig(I) && I->IS_ZOMBIE) ++nzom;        //   count zombie cells <    
  NZOMBIES = nzom;                                 // reset # zombie cells      
}
//------------------------------------------------------------------------------
void builder::de_zombie(box* &B)
  // This routine replaces B with its first non-zombie descendant.              
{
  register indx  i,s;                              // index, counter            
  register node *N,**P;                            // sub-node, octant          
  for(;;) {                                        // for ever                > 
    for(s=0,i=0,N=0,P=B->OCT; i<NSUB; i++,P++)     //   loop octants of box B > 
      if(*P) {                                     //     IF octant active    > 
	if(s) return;                              //       IF not 1st: DONE    
	s = 1;                                     //       ELSE: set s=1       
	N = *P;                                    //       and N=octant        
      }                                            //   < <                     
    B = static_cast<box*>(N);                      //   set B to only sub-node  
  }                                                // <                         
}
//------------------------------------------------------------------------------
#ifdef USE_BODIES
#define SET_SOUL(A,B)    set_soul(A)
#else  // USE_BODIES
#define SET_SOUL(A,B)    set_soul(A,B)
#endif // USE_BODIES
//------------------------------------------------------------------------------
inline indx builder::link_tree(                    // link cells & souls        
                               box * P,            // current box               
                               cell* C,            // current cell              
                               cell*&Cf,           // free cells                
                               soul*&Sf)           // free souls                
  // RECURSIVE
  // This routines transforms the box-leaf tree into the cell-soul tree, such   
  // that all the cells that are contained within some cell are contiguous      
  // in memory, as are the souls.                                               
{
  register indx dep=0;                             // depth of cell             
  C->RADIUS   = P->RADIUS;                         // copy radius               
  C->center() = P->center();                       // copy center               
  C->NUMBER   = P->NUMBER;                         // copy number               
  C->SOULS    = Sf;                                // set cell: sub-souls       
  C->NSOULS   = 0;                                 // reset cell: # sub-souls   
  if(is_twig(P)) {                                 // IF(box==twig)         >   
    C->CELLS  = 0;                                 //   set cell: sub-cells     
    C->NCELLS = 0;                                 //   set cell: # sub-cells   
    register leaf *Li=P->LEAFS;                    //   sub-leaf pointer        
    for(; Li; Li=Li->NEXT) {                       //   loop sub-leafs      >   
      Li->SET_SOUL(Sf++,MASSES);                   //     set soul's body       
      C->NSOULS++;                                 //     increment # sub-souls 
    }                                              //   <                       
  } else {                                         // < ELSE (box==branch)  >   
    register indx   i;                             //   index to octant         
    register node **N;                             //   pointer to current sub  
    de_zombie(P);                                  //   de-zombie               
    for(i=0,N=P->OCT; i<NSUB; i++,N++)             //   loop octants     >      
      if(P->octant_is_leaf(i)) {                   //     IF sub-leaf    >      
        static_cast<leaf*>(*N)->SET_SOUL(Sf++,MASSES); //   set soul's body     
        C->NSOULS++;                               //       inc # sub-souls     
      }                                            //   < <                     
    register indx de,nsub=P->N_subboxes();         //   number of sub-boxes     
    if(nsub) {                                     //   IF sub-boxes         >  
      register box *Pi;                            //     sub-box pointer       
      register cell*Ci=Cf;                         //     remember free cells   
      C->NCELLS = nsub;                            //     set cell: sub-cells   
      C->CELLS  = Cf;                              //     set cell: sub-cells   
      Cf += nsub;                                  //     reserve nsub cells    
      for(i=0,N=P->OCT; i<NSUB; i++,N++)           //     loop octants     >    
        if(*N && P->marked_as_box(i)) {            //       IF sub-box     >    
          Pi = static_cast<box*>(*N);              //         sub-box           
          de = link_tree(Pi,Ci++,Cf,Sf);           //         recursive call    
          if(de>dep) dep=de;                       //         update depth      
        }                                          //     < <                   
      if(C->NSOULS==0) C->SOULS = C->CELLS->SOULS; //     guarantee proper souls
    } else {                                       //   < ELSE (no sub-boxes) > 
      C->NCELLS = 0;                               //     set cell: sub-cells   
      C->CELLS  = 0;                               //     set cell: sub-cells   
    }                                              //   <                       
  }                                                // <                         
  dep++;                                           // increment depth           
  C->DEPTH = dep;                                  // set cell: depth           
  return dep;                                      // return cells depth        
}
#undef SET_SOUL
//------------------------------------------------------------------------------
void builder::link_tree(cell*C0, soul*S0)
{
  register cell* C1=C0+1;
  link_tree(BM->root(),C0,C1,S0);
}
////////////////////////////////////////////////////////////////////////////////
