//-----------------------------------------------------------------------------+
//                                                                             |
// grav.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <tree/grav.h>
#include <tree/stat.h>
#include <iomanip>

////////////////////////////////////////////////////////////////////////////////
#define ISR inline static real
namespace fasttree {
  struct F0 {
    ISR gd0(const real x) { return 1.5-0.5*x; }
    ISR gd1(const real x) { return -one; }
  };
  struct F1 {
    ISR gd0(const real x) { return (0.375*x-1.25)*x+1.875; }
    ISR gd1(const real x) { return 1.5*x-2.5; }
  };
  struct F2 {
    ISR gd0(const real x) { return ((-0.3125*x+1.3125)*x-2.1875)*x+2.1875; }
    ISR gd1(const real x) { return (-1.875*x+5.25)*x-4.375; }
  };
  struct F3 {
    ISR gd0(const real x) { return (((0.2734375*x-1.40625)*x+2.953125)*x
				    -3.28125)*x+2.4609375; }
    ISR gd1(const real x) { return ((2.1875*x-8.4375)*x+11.8125)*x-6.5625; }
  };

#if NDIM==3

  struct K0 {
    ISR gd0(const real x) { return (1.3125*x-3.125)*x+2.8125; }
    ISR gd1(const real x) { return 5.25*x-6.25; }
  };
  struct K1 {
    ISR gd0(const real x) { return ((-1.40625*x+4.59375)*x-5.46875)*x
			           +3.28125; }
    ISR gd1(const real x) { return (-8.4375*x+18.375)*x-10.9375; }
  };
  struct K2 {
    ISR gd0(const real x) { return (((1.50390625*x-6.328125)*x
				     +10.3359375)*x-8.203125)*x+3.69140625; }
    ISR gd1(const real x) { return ((12.03125*x-37.96875)*x+41.34375)*x
			            -16.40625; }
  };
  struct K3 {
    ISR gd0(const real x) { return ((((-1.599609375*x+8.271484375)*x
				      -17.40234375)*x+18.94921875)*x
				    -11.279296875)*x+4.060546875; }
    ISR gd1(const real x) { return (((-15.99609375*x+66.171875)*x
				     -104.4140625)*x+75.796875)*x
			            -22.55859375; }
  };

#else // NDIM==3

  struct K0 {
    ISR gd0(const real x) { return (2.25*x-5)*x+3.75; }
    ISR gd1(const real x) { return 9*x-10; }
  };
  struct K1 {
    ISR gd0(const real x) { return ((-2.5*x+7.875)*x-8.75)*x+4.375; }
    ISR gd1(const real x) { return (-15*x+31.5)*x-17.5; }
  };
  struct K2 {
    ISR gd0(const real x) { return (((2.734375*x-11.25)*x+17.71875)*x
				    -13.125)*x+4.921875; }
    ISR gd1(const real x) { return ((21.875*x-67.5)*x+70.875)*x-26.25; }
  };
  struct K3 {
    ISR gd0(const real x) { return ((((-2.953125*x+15.0390625)*x-30.9375)*x
				     +32.484375)*x-18.046875)*x+5.4140625; }
    ISR gd1(const real x) { return (((-29.53125*x+120.3125)*x-185.625)*x
				    +129.9375)*x-36.09375; }
  };

#endif // NDIM==3
#undef ISR
}

#define P0(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  R *=-D0*x;
  
#define P1(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  register real D1 =-D0*x;						\
  register real heq= half*EPQ;						\
  D0-= heq*D1;								\
  D1+= heq*3*D1*x;							\
  R *= D1;

#define P2(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  register real D1 =-D0*x, D2=-3*D1*x, D3=-5*D2*x;			\
  register real heq= half*EPQ;						\
  D0-= heq*(D1-heq*D2);							\
  D1-= heq*(D2-heq*D3);							\
  R *= D1;

#define P3(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  register real D1 =-D0*x, D2=-3*D1*x, D3=-5*D2*x, D4=-7*D3*x;		\
  register real heq= half*EPQ, hheq=half*heq;				\
  D0 -= ((heq*D3-D2)*hheq+D1)*heq;					\
  D1 -= ((heq*D4-D3)*hheq+D2)*heq;					\
  R  *= D1;
  

#define HARDSOFT_GLOB(MUM,KERNEL)					\
  if(Rq<EPQ) {								\
    x  = Rq*IEQ;							\
    D0 = KERNEL::gd0(x) * (ME=MUM*IEP);					\
    R *= KERNEL::gd1(x) * (ME*IEQ);					\
  } else {								\
    x  = one/Rq;							\
    D0 = MUM*SQRT(x);							\
    R *=-D0*x;								\
  }

#define BODYBODY_GLOB(MUM)						\
switch(KERN) {								\
 case fasttree::f0: { HARDSOFT_GLOB(MUM,F0) } break;			\
 case fasttree::f1: { HARDSOFT_GLOB(MUM,F1) } break;			\
 case fasttree::f2: { HARDSOFT_GLOB(MUM,F2) } break;			\
 case fasttree::f3: { HARDSOFT_GLOB(MUM,F3) } break;			\
 case fasttree::k0: { HARDSOFT_GLOB(MUM,K0) } break;			\
 case fasttree::k1: { HARDSOFT_GLOB(MUM,K1) } break;			\
 case fasttree::k2: { HARDSOFT_GLOB(MUM,K2) } break;			\
 case fasttree::k3: { HARDSOFT_GLOB(MUM,K3) } break;			\
 case fasttree::p0: { P0(MUM)               } break;			\
 case fasttree::p1: { P1(MUM)               } break;			\
 case fasttree::p2: { P2(MUM)               } break;			\
 case fasttree::p3: { P3(MUM)               } break;			\
 default:           { HARDSOFT_GLOB(MUM,F1) } break;			\
}

#define HARDSOFT_INDI(MUM,KERNEL)					\
  EPS= eps_mutual(A,B);							\
  EPQ= EPS*EPS;								\
  if(Rq<EPQ) {								\
    IEP= one/EPS;							\
    IEQ= IEP*IEP;							\
    x  = Rq*IEQ;							\
    D0 = KERNEL::gd0(x) * (ME=MUM*IEP);					\
    R *= KERNEL::gd1(x) * (ME*IEQ);					\
  } else {								\
    x  = one/Rq;							\
    D0 = MUM*SQRT(x);							\
    R *=-D0*x;								\
  }

#define P_INDI(MUM,P_GLOB)						\
  EPS= eps_mutual(A,B);							\
  EPQ= EPS*EPS;								\
  P_GLOB(MUM)

#define BODYBODY_INDI(MUM)						\
switch(KERN) {								\
 case fasttree::f0: { HARDSOFT_INDI(MUM,F0) } break;			\
 case fasttree::f1: { HARDSOFT_INDI(MUM,F1) } break;			\
 case fasttree::f2: { HARDSOFT_INDI(MUM,F2) } break;			\
 case fasttree::f3: { HARDSOFT_INDI(MUM,F3) } break;			\
 case fasttree::k0: { HARDSOFT_INDI(MUM,K0) } break;			\
 case fasttree::k1: { HARDSOFT_INDI(MUM,K1) } break;			\
 case fasttree::k2: { HARDSOFT_INDI(MUM,K2) } break;			\
 case fasttree::k3: { HARDSOFT_INDI(MUM,K3) } break;			\
 case fasttree::p0: { P_INDI(MUM,P0)        } break;			\
 case fasttree::p1: { P_INDI(MUM,P1)        } break;			\
 case fasttree::p2: { P_INDI(MUM,P2)        } break;			\
 case fasttree::p3: { P_INDI(MUM,P3)        } break;			\
 default:           { HARDSOFT_INDI(MUM,F1) } break;			\
}

#define GREENS_FUNCTION							\
  register real	x,D0,D1,D2,D3;						\
  switch(KERN) {							\
  case fasttree::p3: {							\
    register real D4,D5,D6,heq=half*EPQ,hheq=half*heq;			\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =   -D0*x;							\
    D2 = -3*D1*x;							\
    D3 = -5*D2*x;							\
    D4 = -7*D3*x,							\
    D5 = -9*D4*x;							\
    D6 =-11*D5*x;							\
    D0-= ((heq*D3-D2)*hheq+D1)*heq;					\
    D1-= ((heq*D4-D3)*hheq+D2)*heq;					\
    D2-= ((heq*D5-D4)*hheq+D3)*heq;					\
    D3-= ((heq*D6-D5)*hheq+D4)*heq;					\
    } break;								\
  case fasttree::p2: {							\
    register real D4,D5,heq=half*EPQ;					\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
    D4 =-7*D3*x,							\
    D5 =-9*D4*x;							\
    D0-= heq*(D1-heq*D2);						\
    D1-= heq*(D2-heq*D3);						\
    D2-= heq*(D3-heq*D4);						\
    D3-= heq*(D4-heq*D5);						\
    } break;								\
  case fasttree::p1: {							\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
    register real heq=half*EPQ;						\
    D0 -= heq*D1;							\
    D1 -= heq*D2;							\
    D2 -= heq*D3;							\
    D3 += heq*7*D3*x;							\
    } break;								\
  case fasttree::p0:							\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
    break;								\
  default:								\
    x  = one/Rq;							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
  }

////////////////////////////////////////////////////////////////////////////////
using namespace fasttree;

////////////////////////////////////////////////////////////////////////////////
// constants that determine the behaviour of the gravity routines               

#if NDIM==3
const fasttree::uint N_PRE [3] = { 3, 0,12};
const fasttree::uint N_POST[3] = {64,12, 0};  // N_POST[2] is not checked
#else
const fasttree::uint N_PRE [3] = { 3, 0,12};
const fasttree::uint N_POST[3] = {64, 8, 0};  // N_POST[2] is not checked
#endif

////////////////////////////////////////////////////////////////////////////////
// For individual softening:                                                    
// The little inline function eps_mutual determines the mutual softening        
// between two bodies. If softened cell interactions are enabled, then          
// there are also routines for eps_mutual between cells and cell and body.      
// Otherwise, there are routines for the minimum and maximum eps between        
// two cells and cell and body.                                                 

namespace fasttree {
#ifdef SOFT_MIN      // eps_mutual = min (eps_i, eps_j)

  inline real eps_mutual(soul*A, soul*B) { return min(eps(A),eps(B)); }
  inline real eps_mutual(cell*A, soul*B) { return min(emin(A),eps(B)); }
  inline real eps_mutual(cell*A, cell*B) { return min(emin(A),emin(B)); }

#else  // SOFT_MIN      eps_mutual = mass-weighted mean of eps_i and eps_j

  inline real eps_mutual(soul*A, soul*B)
  {
    if(eps(A)  == eps(B))  return eps(A);
    if(mass(A) == mass(B)) return half*(eps(A)+eps(B));
    return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B));
  }
  inline real eps_mutual(cell*A, soul*B)
  { if(eps(A)==eps(B)) return eps(A);
  else               return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B)); }
  inline real eps_mutual(cell*A, cell*B)
  { if(eps(A)==eps(B)) return eps(A);
  else               return (mass(A)*eps(A)+mass(B)*eps(B))/(mass(A)+mass(B)); }

#endif // SOFT_MIN

  inline real emin(const cell *a, const cell *b) { return min(emin(a),emin(b));}
  inline real emin(const cell *a, const soul *b) { return min(emin(a),eps(b)); }
  inline real emax(const cell *a, const cell *b) { return max(emax(a),emax(b));}
  inline real emax(const cell *a, const soul *b) { return max(emax(a),eps(b)); }
}
////////////////////////////////////////////////////////////////////////////////
// class gravity_kernel                                                         
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
// body-many bodies interaction for global and individual softening             
//------------------------------------------------------------------------------
#define LOOP_B for(register soul*B=B0; B!=BN; B++)
#ifdef TESTBODIES
// macro LOOPMANY performs the body-body interactions between body A and the    
// bodies from B0 to BN. There are two versions, one which allows test bodies   
// and one which does not (default). The macros SET_RAD, MAMB, SET_AB, etc      
// determine the details and MUST be defined somewhere else (see below).        
#define LOOPMANY								\
if(A_is_sink && A_is_source)							\
  switch(B_sink) {								\
  case all:									\
    switch(B_source) {								\
    case all:  LOOP_B {                    SET_RAD MAMB SET_AB } break;		\
    case some: LOOP_B { if(is_source(B)) { SET_RAD MAMB SET_AB }		\
                        else             { SET_RAD MMA  SET_B  }		\
                      } break;							\
    case none: LOOP_B {                    SET_RAD MMA  SET_B  } break;		\
    } break;									\
  case some:									\
    switch(B_source) {								\
    case all:  LOOP_B { if(is_sink(B)) { SET_RAD MAMB SET_AB }			\
	                else           { SET_RAD MAMB SET_A  }			\
                      } break;							\
    case some: LOOP_B {	if(is_source(B) && is_sink(B)) { SET_RAD MAMB SET_AB }	\
	                else if(is_source(B))          { SET_RAD MAMB SET_A  }	\
	                else if(is_sink(B))            { SET_RAD MMA  SET_B  }	\
                      } break;							\
    case none: LOOP_B { if(is_sink(B)) { SET_RAD MMA SET_B } } break;		\
    } break;									\
  case none:									\
    switch(B_source) {								\
    case all:  LOOP_B {                    SET_RAD MAMB SET_A   } break;	\
    case some: LOOP_B { if(is_source(B)) { SET_RAD MAMB SET_A } } break;	\
    case none: break;								\
    } break;									\
  }										\
else if(A_is_source)								\
  switch(B_sink) {								\
  case all:									\
    switch(B_source) {								\
    case all:  LOOP_B {                    SET_RAD MAMB SET_B } break;		\
    case some: LOOP_B { if(is_source(B)) { SET_RAD MAMB SET_B }			\
                        else             { SET_RAD MMA  SET_B }			\
                      } break;							\
    case none: LOOP_B {                    SET_RAD MMA  SET_B  } break;		\
    } break;									\
  case some:									\
    switch(B_source) {								\
    case all:  LOOP_B { if(is_sink(B)) { SET_RAD MAMB SET_B } } break;		\
    case some: LOOP_B { if(is_sink(B)) {					\
                          if(is_source(B)) { SET_RAD MAMB SET_B }		\
                          else             { SET_RAD MMA  SET_B  }		\
                      } } break;						\
    case none: LOOP_B { if(is_sink(B)) { SET_RAD MMA SET_B } } break;		\
    } break;									\
  case none: break;								\
  }										\
else if(A_is_sink)								\
  switch(B_source) {								\
  case all:  LOOP_B {                    SET_RAD MMB SET_A   } break;		\
  case some: LOOP_B { if(is_source(B)) { SET_RAD MMB SET_A } } break;		\
  case none: break;								\
  }
#else  // TESTBODIES
#define LOOPMANY								\
if(A_is_sink)									\
  switch(B_sink) {								\
  case all:  LOOP_B { SET_RAD MAMB SET_AB } break;				\
  case some: LOOP_B { if(is_sink(B)) { SET_RAD MAMB SET_AB }			\
                      else           { SET_RAD MAMB SET_A  } } break;		\
  case none: LOOP_B { SET_RAD MAMB SET_A } break;				\
  }										\
else										\
  switch(B_sink) {								\
  case all:  LOOP_B { SET_RAD MAMB SET_B } break;				\
  case some: LOOP_B if(is_sink(B)) { SET_RAD MAMB SET_B } break;		\
  case none: break;								\
  }
#endif // TESTBODIES
//------------------------------------------------------------------------------
// define the following macros, such that macro LOOPMANY above results in the   
// computation of body-body interactions with global softening length.          
#define SET_RAD   R  = cofm(A)-cofm(B); Rq = norm(R);
#define MAMB      MM = mass(A)*mass(B);
#define MMA       MM = mass(A);
#define MMB       MM = mass(B);
#define GRAV_A    A->pot()-= D0; A->acc()+= R;
#define GRAV_B    B->pot()-= D0; B->acc()-= R;
#define SET_AB    { BODYBODY_GLOB(MM) GRAV_A GRAV_B }
#define SET_A     { BODYBODY_GLOB(MM) GRAV_A }
#define SET_B     { BODYBODY_GLOB(MM) GRAV_B }

void gravity_kernel::many_global(const bool A_is_sink,
#ifdef TESTBODIES
				 const bool A_is_source,
#endif
				 soul*A,
				 const who B_sink,
#ifdef TESTBODIES
				 const who B_source,
#endif
				 soul*B0, const soul*BN) const
  // This routine performs body-body interactions between body A and the        
  // bodies from B0 to BN. Using a fixed softening length.                      
{
  register real Rq,x,D0,MM,ME;
  register vect R;
  LOOPMANY
}
//------------------------------------------------------------------------------
#undef SET_AB
#undef SET_A
#undef SET_B
// define the following macros, such that macro LOOPMANY above results in the   
// computation of body-body interactions with individual softening lengths.     
#define SET_AB    { BODYBODY_INDI(MM) GRAV_A GRAV_B }
#define SET_A     { BODYBODY_INDI(MM) GRAV_A }
#define SET_B     { BODYBODY_INDI(MM) GRAV_B }
void gravity_kernel::many_individual(const bool A_is_sink,
#ifdef TESTBODIES
				     const bool A_is_source,
#endif
				     soul*A,
				     const who B_sink,
#ifdef TESTBODIES
				     const who B_source,
#endif
				     soul*B0, const soul*BN) const
  // This routine performs body-body interactions between body A and the        
  // bodies from B0 to BN. Using individual softening lengths.                  
{
  register real Rq,x,D0,MM,ME;
  register vect R;
  LOOPMANY
}
#undef  LOOPMANY
//------------------------------------------------------------------------------
inline gravity_kernel::who gravity_kernel::who_is_sink(const cell* C) const
  // This routines figures out whether all, none, or some of the bodies in cell 
  // C are sinks.                                                               
{
  if(!is_sink(C)) return none;
  if( al_sink(C)) return all;
                  return some;
}
//------------------------------------------------------------------------------
#ifdef TESTBODIES
inline gravity_kernel::who gravity_kernel::who_is_source(const cell* C) const
  // This routines figures out whether all, none, or some of the bodies in cell 
  // C are sources.                                                             
{
  if(!is_source(C)) return none;
  if( al_source(C)) return all;
                    return some;
}
//------------------------------------------------------------------------------
// replacing single cell-body interaction with body-many body interactions      
//------------------------------------------------------------------------------
inline void gravity_kernel::many_global(const bool is_sink_A,
					const bool is_source_A,
					soul* A, cell* B) const
{
  many_global(is_sink_A,is_source_A,A,
	      who_is_sink(B),who_is_source(B),
	      B->begin_allsouls(),B->end_allsouls());
}
inline void gravity_kernel::many_individual(const bool is_sink_A,
					    const bool is_source_A,
					    soul* A, cell* B) const
{
  many_individual(is_sink_A,is_source_A,A,
		  who_is_sink(B),who_is_source(B),
		  B->begin_allsouls(),B->end_allsouls());
}
//------------------------------------------------------------------------------
inline void gravity_kernel::many_global(cell* A, soul* B) const
{
  many_global(is_sink(B),is_source(B),B,
	      who_is_sink(A),who_is_source(A),
	      A->begin_allsouls(),A->end_allsouls());
}
inline void gravity_kernel::many_individual(cell* A, soul* B) const
{
  many_individual(is_sink(B),is_source(B),B,
		  who_is_sink(A),who_is_source(A),
		  A->begin_allsouls(),A->end_allsouls());
}
#else  // TESTBODIES============================================================
inline void gravity_kernel::many_global(cell* A, soul* B) const
{
  many_global(is_sink(B),B,who_is_sink(A),A->begin_allsouls(),A->end_allsouls());
}
inline void gravity_kernel::many_individual(cell* A, soul* B) const
{
  many_individual(is_sink(B),B,who_is_sink(A),
		  A->begin_allsouls(),A->end_allsouls());
}
#endif // TESTBODIES
//------------------------------------------------------------------------------
// replacing single cell-cell interaction with body-many body interactions      
//------------------------------------------------------------------------------
#define LoopAllSouls(C,S) for(register cell::allsoul_iterator			 \
			       S=C->begin_allsouls(); S!=C->end_allsouls(); ++S)
//------------------------------------------------------------------------------
inline void gravity_kernel::many_global(cell* A, cell* B) const
{
#ifdef TESTBODIES
  if(al_sink(A)) {
    if(al_source(A))
      LoopAllSouls(A,Ai)       many_global(true,true,Ai,B);
    else if(is_source(A))
      LoopAllSouls(A,Ai)       many_global(true,is_source(Ai),Ai,B);
    else
      LoopAllSouls(A,Ai)       many_global(true,false,Ai,B);
  } else if(is_sink(A)) {
    if(al_source(A))
      LoopAllSouls(A,Ai)       many_global(is_sink(Ai),true,Ai,B);
    else if(is_source(A)) {
      LoopAllSouls(A,Ai)
        if(is_sink(Ai))        many_global(true,is_source(Ai),Ai,B);
        else if(is_source(Ai)) many_global(false,true,Ai,B);
    } else
      LoopAllSouls(A,Ai)
	if(is_sink(Ai))        many_global(true,false,Ai,B);
  } else {
    if(al_source(A))
      LoopAllSouls(A,Ai)       many_global(false,true,Ai,B);
    else if(is_source(A))
      LoopAllSouls(A,Ai)
	if(is_source(Ai))      many_global(false,true,Ai,B);
  }
#else  // TESTBODIES
  if(al_sink(A))
    LoopAllSouls(A,Ai)
      many_global(true,Ai,who_is_sink(B),B->begin_allsouls(),B->end_allsouls());
  else if(is_sink(A))
    LoopAllSouls(A,Ai)
      many_global(is_sink(Ai),Ai,who_is_sink(B),B->begin_allsouls(),B->end_allsouls());
  else if(is_sink(B))
    LoopAllSouls(A,Ai)
      many_global(false,Ai,who_is_sink(B),B->begin_allsouls(),B->end_allsouls());
#endif // TESTBODIES
}
//------------------------------------------------------------------------------
inline void gravity_kernel::many_individual(cell* A, cell* B) const
{
#ifdef TESTBODIES
  if(al_sink(A)) {
    if(al_source(A))
      LoopAllSouls(A,Ai)       many_individual(true,true,Ai,B);
    else if(is_source(A))
      LoopAllSouls(A,Ai)       many_individual(true,is_source(Ai),Ai,B);
    else
      LoopAllSouls(A,Ai)       many_individual(true,false,Ai,B);
  } else if(is_sink(A)) {
    if(al_source(A))
      LoopAllSouls(A,Ai)       many_individual(is_sink(Ai),true,Ai,B);
    else if(is_source(A)) {
      LoopAllSouls(A,Ai)
	if(is_sink(Ai))        many_individual(true,is_source(Ai),Ai,B);
	else if(is_source(Ai)) many_individual(false,true,Ai,B);
    } else
      LoopAllSouls(A,Ai)
	if(is_sink(Ai))        many_individual(true,false,Ai,B);
  } else {
    if(al_source(A))
      LoopAllSouls(A,Ai)       many_individual(false,true,Ai,B);
    else if(is_source(A))
      LoopAllSouls(A,Ai)
	if(is_source(Ai))      many_individual(false,true,Ai,B);
  }
#else  // TESTBODIES
  if(al_sink(A)) {
    LoopAllSouls(A,Ai)
      many_individual(true,Ai,who_is_sink(B),B->begin_allsouls(),B->end_allsouls());
  } else if(is_sink(A)) {
    LoopAllSouls(A,Ai)
      many_individual(is_sink(Ai),Ai,who_is_sink(B),B->begin_allsouls(),B->end_allsouls());
  } else if(is_sink(B)) {
    LoopAllSouls(A,Ai) 
      many_individual(false,Ai,who_is_sink(B),B->begin_allsouls(),B->end_allsouls());
  }
#endif // TESTBODIES
}
//------------------------------------------------------------------------------
// replacing single cell self-interaction with body-many body interactions      
//------------------------------------------------------------------------------
#undef  LoopAllSouls
#define LoopAllSouls(C,S) for(register cell::allsoul_iterator			 \
			  S=C->begin_allsouls(); S!=C->last_allsouls(); ++S)
//------------------------------------------------------------------------------
inline void gravity_kernel::many_global(cell* A) const
{
#ifdef TESTBODIES
  if(al_sink(A)) {
    if(al_source(A))
      LoopAllSouls(A,Ai)
	many_global(true,true,Ai,all,all,Ai+1,A->end_allsouls());
    else
      LoopAllSouls(A,Ai)
	many_global(true,is_source(Ai),Ai,all,some,Ai+1,A->end_allsouls());
  } else {
    if(al_source(A))
      LoopAllSouls(A,Ai)
	many_global(is_sink(Ai),true,Ai,some,all,Ai+1,A->end_allsouls());
    else if(is_source(A))
      LoopAllSouls(A,Ai)
	many_global(is_sink(Ai),is_source(Ai),Ai,some,some,Ai+1,A->end_allsouls());
  }
#else  // TESTBODIES
  if(al_sink(A)) {
    LoopAllSouls(A,Ai)
      many_global(true,Ai,all,Ai+1,A->end_allsouls());
  } else {
    LoopAllSouls(A,Ai)
      many_global(is_sink(Ai),Ai,some,Ai+1,A->end_allsouls());
  }
#endif // TESTBODIES
}
//------------------------------------------------------------------------------
inline void gravity_kernel::many_individual(cell* A) const
{
#ifdef TESTBODIES
  if(al_sink(A)) {
    if(al_source(A))
      LoopAllSouls(A,Ai)
	many_individual(true,true,Ai,all,all,Ai+1,A->end_allsouls());
    else 
      LoopAllSouls(A,Ai)
	many_individual(true,is_source(Ai),Ai,all,some,Ai+1,A->end_allsouls());
  } else {
    if(al_source(A))
      LoopAllSouls(A,Ai)
	many_individual(is_sink(Ai),true,Ai,some,all,Ai+1,A->end_allsouls());
    else 
      LoopAllSouls(A,Ai)
	many_individual(is_sink(Ai),is_source(Ai),Ai,some,some,Ai+1,
			A->end_allsouls());
  }
#else  // TESTBODIES
  if(al_sink(A)) {
    LoopAllSouls(A,Ai)
      many_individual(true,Ai,all,Ai+1,A->end_allsouls());
  } else {
    LoopAllSouls(A,Ai)
      many_individual(is_sink(Ai),Ai,some,Ai+1,A->end_allsouls());
  }
#endif // TESTBODIES
}
//------------------------------------------------------------------------------
template<class NA, class NB>
inline real mutual_mass(const NA*A, const NB*B)
  // This routines usually returns the product of the masses of nodes A and B.  
  // only in case test bodies are allowed AND one of the nodes is a test node,  
  // i.e. is not a source, its mass is not multiplied in.                       
{
#ifdef TESTBODIES
  if(is_source(A))
    if(is_source(B)) return mass(A)*mass(B);
    else             return mass(A);
  else
    if(is_source(B)) return mass(B);
    else             return one;
#else
  return mass(A)*mass(B);
#endif
}    
//------------------------------------------------------------------------------
template<class NA, class NB>
inline bool interacting(const NA*A, const NB*B)
  // This routines figures out whether the interaction  A <- B  is needed.      
{
#ifdef TESTBODIES
  return is_sink(A) && is_source(B);               // A must be sink & B source 
#else
  return is_sink(A);                               // A must be sink            
#endif
}
//------------------------------------------------------------------------------
inline void gravity_kernel::grav(vect &R, const real Rq, cell *A, soul *B) const
  // This routines performs a mutual cell-body interaction.                     
{
  GREENS_FUNCTION                                  // define Dn (see kern.h)    
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    c0(A)+= D0;                                    //   add to C0 of A          
    c1(A).add_mul(R, D1);                          //   add to C1 of A          
    c2(A).add_quadrup(R,D1,D2);                    //   add to C2 of A          
    c3(A).add_quadrup(R,D2,D3);                    //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    register vect RQ = R * quadru(A);		   //   R*Q_A                   
    register real tQ = trace(quadru(A)), RQR=R*RQ; //   tr(Q_A), R*Q*R          
    B->pot()-=     D0+half*(D1*tQ + D2*RQR);	   //   pot A->B                
    B->acc().sub_mul(R, D1+half*(D2*tQ + D3*RQR)); //   acc A->B (term in R)    
    B->acc().sub_mul(RQ,D2);                       //   acc A->B (term in R*Q_A)
  }                                                // <                         
}
//------------------------------------------------------------------------------
inline void gravity_kernel::grav(vect &R,const real Rq, cell* A,cell *B) const
  // This routines performs a mutual cell-cell interaction.                     
  // It's at the root of the whole code; most of the CPU time is spent in this  
  // routine and its (inlined) calls.                                           
{
  GREENS_FUNCTION                                  // define Dn (see kern.h)    
  register matr C2; C2.quadrup(R,D1,D2);           // pre-compute C2            
  register tens C3; C3.quadrup(R,D2,D3);           // pre-compute C3            
  register vect RQ;                                // to hold R*Q               
  register real tQ, RQR;                           // to hold tr(Q), R*Q*R      
  if(interacting(A,B)) {                           // IF(interaction B->A) >    
    RQ    = R*quadru(B);                           //   R*Q_B                   
    tQ    = trace(quadru(B));                      //   tr(Q_B)                 
    RQR   = R*RQ;                                  //   R*Q_B*R                 
    c0(A)+= D0 +half*(D1*tQ + D2*RQR);             //   add to C0 of A          
    c1(A).add_mul(R, D1 +half*(D2*tQ + D3*RQR));   //   add to C1 of A (R)      
    c1(A).add_mul(RQ,D2);                          //   add to C1 of A (R*Q_B)  
    c2(A)+= C2;                                    //   add to C2 of A          
    c3(A)+= C3;                                    //   add to C3 of A          
  }                                                // <                         
  if(interacting(B,A)) {                           // IF(interaction A->B) >    
    R.negate();                                    //   set R to -R             
    RQ    = R*quadru(A);                           //   R*Q_A                   
    tQ    = trace(quadru(A));                      //   tr(Q_A)                 
    RQR   = R*RQ;                                  //   R*Q_A*R                 
    c0(B)+= D0 +half*(D1*tQ + D2*RQR);             //   add to C0 of B          
    c1(B).add_mul(R, D1 +half*(D2*tQ + D3*RQR));   //   add to C1 of A (R)      
    c1(B).add_mul(RQ,D2);                          //   add to C1 of A (R*Q_B)  
    c2(B)+= C2;                                    //   add to C2 of B          
    c3(B)-= C3;                                    //   add to C3 of B          
  }                                                // <                         
}
//------------------------------------------------------------------------------
namespace fasttree {
  inline bool self_softened(const cell   *a,       // I:  sink = source         
			    const real    eps)     // I:  softening length      
    // This routine figures out if a cell-self interaction is fully-softened.   
  {
    return eps > (rmax(a)+rmax(a));
  }
  //----------------------------------------------------------------------------
  inline bool fully_softened(const cell   *a,      // I:  cell A                
			     const cell   *b,      // I:  cell B                
			     const real    Rq,     // I:  |x-y|^2               
			     const real    eps)    // I:  softening length      
    // This routine figures out if a cell-cell interaction is fully softened.   
  {
    register real x = eps - rmax(a) - rmax(b);
    return x>zero && Rq<(x*x);
  }
  //----------------------------------------------------------------------------
  inline bool fully_softened(const cell   *b,      // I:  cell B                
			     const real    Rq,     // I:  |x-y|^2               
			     const real    eps)    // I:  softening length      
    // This routine figures out if a cell-body interaction is fully softened.   
  {
    return eps>rmax(b) && Rq<square(eps-rmax(b));
  }
  //----------------------------------------------------------------------------
  inline bool fully_newtonian(const cell   *a,     // I:  cell A                
			      const cell   *b,     // I:  cell B                
			      const real    Rq,    // I:  |x-y|^2               
			      const real    eps)   // I:  softening length      
    // This routine figures out if a cell-cell interaction is fully Newtonian.  
  {
    return square(eps+rmax(a)+rmax(b)) < Rq;
  }
  //----------------------------------------------------------------------------
  inline bool fully_newtonian(const cell   *b,     // I:  cell B                
			      const real    Rq,    // I:  |x-y|^2               
			      const real    eps)   // I:  softening length      
    // This routine figures out if a cell-body interaction is fully Newtonian.  
  {
    return square(eps+rmax(b)) < Rq;
  }
  //----------------------------------------------------------------------------
  inline bool split_into_many(const cell   *a,     // I:  cell A                
			      const cell   *b,     // I:  cell B                
			      const uint    Ni)    // I:  min # body-body ias   
    // This routine figures out if a cell-cell interaction can be done directly.
  {
    return 4u < Ni && number(a)*number(b) < Ni;
  }
  //----------------------------------------------------------------------------
  inline bool split_into_many(const cell   *b,     // I:  cell                  
			      const uint    Ni)    // I:  min # body-body ias   
    // This routine figures out if a cell-self or a cell-body interaction shall 
    // be done directly.                                                        
  {
    return 2u < Ni && number(b) < Ni;
  }
  //----------------------------------------------------------------------------
  inline bool well_separated(const cell   *a,      // I:  cell A                
			     const cell   *b,      // I:  cell B                
			     const real    Rq)     // I:  |x-y|^2               
    // This routine finds out whether a cell-cell interaction is well-separated.
  {
    return Rq > square(rcrit(a)+rcrit(b));
  }
  //----------------------------------------------------------------------------
  inline bool well_separated(const cell   *b,      // I:  cell                  
			     const real    Rq)     // I:  |x-y|^2               
    // This routine finds out whether a cell-body interaction is well-separated.
  {
    return Rq > rcrit2(b);
  }
  //----------------------------------------------------------------------------
  inline bool are_twigs(const cell *a,             // I:   cell A               
			const cell *b)             // I:   cell B               
    // This routine figures out whether both cells are twig cells.              
  {
    return is_twig(a) && is_twig(b);
  }
  //----------------------------------------------------------------------------
  template<class NA, class NB>
  inline bool have_interaction(const NA*A, const NB*B)
    // This routines figures out whether any interaction is needed between two  
    // nodes A and B.                                                           
  { 
    return interacting(A,B) || interacting(B,A);
  }
}
//------------------------------------------------------------------------------
void gravity_kernel::interact(soul*A, soul*B) const
  // This routines performs an elementary body-body interaction.                
{
  if(!have_interaction(A,B)) return;               // IF no interaction > DONE <
  register vect R  = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(R),x,D0,ME;              // compute R^2               
  switch(SOFT) {                                   // what kind of softening? > 
  case individual:                                 // - individual softening >  
    BODYBODY_INDI(mutual_mass(A,B)) break;         //   compute (see kern.h) <  
  default:                                         // - global softening     >  
    BODYBODY_GLOB(mutual_mass(A,B)) break;         //   compute (see kern.h) <  
  }                                                // <                         
  if(interacting(A,B)) { A->pot()-=D0; A->acc()+=R; }// IF(B->A) > add to A <   
  if(interacting(B,A)) { B->pot()-=D0; B->acc()-=R; }// IF(A->B) > add to B <   
  STAT->record_BB();                               // record statistics         
}  
//==============================================================================
inline bool gravity_kernel::mutual_global(cell*A, soul*B) const
{
  if(split_into_many(A,N_PRE[0])) {                // IF(make many BBs) >       
    many_global(A,B);                              //   perform BB iactions     
    STAT->record_MB();                             //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(dX);                     // compute R^2               
  if(well_separated(A,Rq) &&                       // IF(well separated         
     (continuous_kernel()  ||                      // AND(continuous g(r)       
      fully_newtonian(A,Rq,EPS))) {                //     OR fully Newtonian)) >
    grav(dX,Rq,A,B);                               //   interact                
    STAT->record_CB();                             //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  if(is_twig(A) ||                                 // IF(twig cell              
     split_into_many(A,N_POST[0]) ||               // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,Rq,EPS))) {                 //    AND fully softened)) > 
    many_global(A,B);                              //   perform BB iactions     
    STAT->record_MB();                             //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  return false;                                    // SPLIT                     
}
//------------------------------------------------------------------------------
inline bool gravity_kernel::mutual_individual(cell*A, soul*B) const
{
  if(split_into_many(A,N_PRE[0])) {                // IF(make many BBs) >       
    many_individual(A,B);                          //   perform BB iactions     
    STAT->record_MB();                             //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute R = x_A-x_B       
  register real Rq = norm(dX);                     // compute R^2               
  if(well_separated(A,Rq) &&                       // IF(well separated         
     (continuous_kernel()  ||                      // AND (continuous g(r)      
      fully_newtonian(A,Rq,emax(A,B)))) {          //     OR fully Newtonian)) >
    if(continuous_kernel()) {                      //   IF(continuous g(r)) >   
      EPS = eps_mutual(A,B);                       //     set eps               
      EPQ = EPS*EPS;                               //     and eps^2             
    }                                              //   <                       
    grav(dX,Rq,A,B);                               //   interact                
    STAT->record_CB();                             //   record statistics       
    return true;                                   //   DONE                    
  }                                                // <                         
  if(is_twig(A) ||                                 // IF(twig cell              
     split_into_many(A,N_POST[0]) ||               // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,Rq,emin(A,B)))) {           //    AND fully softened)) > 
    many_individual(A,B);                          //   perform BB iactions     
    STAT->record_MB();                             //   record stats            
    return true;                                   //   DONE                    
  }                                                // <                         
  return false;                                    // SPLIT                     
}
//------------------------------------------------------------------------------
bool gravity_kernel::interact(cell*A, soul*B) const
{
  if(!have_interaction(A,B)) return true;
  switch(SOFT) {
  case individual: return mutual_individual(A,B);
  default:               return mutual_global(A,B);
  }
}
//==============================================================================
inline bool gravity_kernel::mutual_global(cell*A, cell*B) const
{
  if(split_into_many(A,B,N_PRE[1])) {              // IF make many BBs       >  
    many_global(A,B);                              //    perform BB iactions    
    STAT->record_MB();                             //    record stats           
    return true;                                   //    DONE                   
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute dX = X_A - X_B    
  register real Rq = norm(dX);                     // and dX^2                  
  if(well_separated (A,B,Rq) &&                    // IF well separated         
     (continuous_kernel() ||                       // AND    continuous g(r)    
      fully_newtonian(A,B,Rq,EPS))) {              //     OR fully Newtonian  > 
    grav(dX,Rq,A,B);                               //    interact               
    STAT->record_CC();                             //    record stats           
    return true;                                   //    DONE                   
  }                                                // <                         
  if(are_twigs(A,B) ||                             // IF twig cells             
     split_into_many(A,B,N_POST[1]) ||             // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,B,Rq,EPS))) {               //    AND fully softened)) > 
    many_global(A,B);                              //    perform BB iactions    
    STAT->record_MB();                             //    record stats           
    return true;                                   //    DONE                   
  }                                                // < ELSE  >                 
  return false;                                    //   SPLIT <                 
}
//------------------------------------------------------------------------------
inline bool gravity_kernel::mutual_individual(cell*A, cell*B) const
{
  if(split_into_many(A,B,N_PRE[1])) {              // IF make many BBs       >  
    many_individual(A,B);                          //    perform BB iactions    
    STAT->record_MB();                             //    record stats           
    return true;                                   //    DONE                   
  }                                                // <                         
  register vect dX = cofm(A)-cofm(B);              // compute dX = X_A - X_B    
  register real Rq = norm(dX);                     // and dX^2                  
  if(well_separated (A,B,Rq) &&                    // IF  well separated        
     (continuous_kernel() ||                       // AND    continuous g(r)    
      fully_newtonian(A,B,Rq,emax(A,B)))) {        //     OR fully Newtonian >  
    if(continuous_kernel()) {                      //    IF continuous g(r)    >
      EPS = eps_mutual(A,B);                       //       set eps             
      EPQ = EPS*EPS;                               //       and eps^2           
    }                                              //    <                      
    grav(dX,Rq,A,B);                               //    interact hard          
    STAT->record_CC();                             //    record stats           
    return true;                                   //    DONE                   
  }                                                // <                         
  if(are_twigs(A,B) ||                             // IF twig cells             
     split_into_many(A,B,N_POST[1]) ||             // OR make many BBs          
     (!continuous_kernel() &&                      // OR(non-continuous g(r)    
      fully_softened(A,B,Rq,emin(A,B)))) {         //    AND fully softened)) > 
    many_individual(A,B);                          //    perform BB iactions    
    STAT->record_MB();                             //    record stats           
    return true;                                   //    DONE                   
  }                                                // < ELSE  >                 
  return false;                                    //   SPLIT <                 
}
//------------------------------------------------------------------------------
bool gravity_kernel::interact(cell*A, cell*B) const
{
  if(!have_interaction(A,B)) return true;
  switch(SOFT) {
  case individual: return mutual_individual(A,B);
  default:               return mutual_global(A,B);
  }
}
//==============================================================================
bool gravity_kernel::interact(cell*A) const
{
#ifdef TESTBODIES
  if(!is_sink(A) || !is_source(A)) return true;
#else
  if(!is_sink(A)) return true;
#endif
  switch(SOFT){
  case individual:
    if(is_twig(A) ||
       split_into_many(A,N_PRE[2]) ||
       !continuous_kernel() && self_softened(A,emin(A))) { 
      many_individual(A);
      STAT->record_MB();
      return true;
    }
    return false;
  default:
    if(is_twig(A) ||
       split_into_many(A,N_PRE[2]) ||
       !continuous_kernel() && self_softened(A,EPS)) {
      many_global(A);
      STAT->record_MB();
      return true;
    }
    return false;
  }
}
//------------------------------------------------------------------------------
