/*******************************************************************************
 *                                                                             *
 *  treeF.c                                                                    *
 *                                                                             *
 *  written by Walter Dehnen, 2000/2001                                        *
 *  e-mail:   dehnen@mpia-hd.mpg.de                                            *
 *  address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg  *
 *            Germany                                                          *
 *                                                                             *
 ******************************************************************************/

#include <treeC.h>

#include <time.h>
#include <stdlib.h>

#if NDIM==3
#define POSIS   const INPUT_TYPE*X, const INPUT_TYPE*Y, const INPUT_TYPE*Z
#define POSS    X,Y,Z
#define VELIS   const INPUT_TYPE*VX, const INPUT_TYPE*VY, const INPUT_TYPE*VZ
#define VELS    VX,VY,VZ
#define ACCIS   INPUT_TYPE*AX, INPUT_TYPE*AY, INPUT_TYPE*AZ
#define ACCS    AX,AY,AZ
#define ZEROS   0,0,0
#else
#define POSIS   const INPUT_TYPE*X, const INPUT_TYPE*Y
#define POSS    X,Y
#define VELIS   const INPUT_TYPE*VX, const INPUT_TYPE*VY
#define VELS    VX,VY
#define ACCIS   INPUT_TYPE*AX, INPUT_TYPE*AY
#define ACCS    AX,AY
#define ZEROS   0,0
#endif

/******************************************************************************/
void treef_init_ (int *F, INPUT_TYPE*M, POSIS, INPUT_TYPE*E, ACCIS,
		  INPUT_TYPE*PH, INPUT_TYPE*RH, int*N, INPUT_TYPE*EPS,
		  INPUT_TYPE*TH, int*K, int*DMAX)
{ TreeC_initialize(F,M,POSS,E,ACCS,PH,RH,*N,*EPS,*TH,*K,*DMAX); }
void treef_init__(int *F, INPUT_TYPE*M, POSIS, INPUT_TYPE*E, ACCIS,
		  INPUT_TYPE*PH, INPUT_TYPE*RH, int*N, INPUT_TYPE*EPS,
		  INPUT_TYPE*TH, int*K, int*DMAX)
{ TreeC_initialize(F,M,POSS,E,ACCS,PH,RH,*N,*EPS,*TH,*K,*DMAX); }
/******************************************************************************/
void treef_resetsoftening_ (INPUT_TYPE*EPS, int*K)
                                             { TreeC_resetsoftening(*EPS,*K); }
void treef_resetsoftening__(INPUT_TYPE*EPS, int*K)
                                             { TreeC_resetsoftening(*EPS,*K); }
/******************************************************************************/
void treef_resetopening_ (INPUT_TYPE*TH)     { TreeC_resetopening(*TH); }
void treef_resetopening__(INPUT_TYPE*TH)     { TreeC_resetopening(*TH); }
/******************************************************************************/
void treef_clearup_ ()                       { TreeC_clearup(); }
void treef_clearup__()                       { TreeC_clearup(); }
/******************************************************************************/
void treef_grow_ (int*nx)                    { TreeC_grow(*nx); }
void treef_grow__(int*nx)                    { TreeC_grow(*nx); }
/******************************************************************************/
void treef_grow_centered_ (int*nx,int X[NDIM]) { TreeC_grow_centered(*nx,X); }
void treef_grow_centered__(int*nx,int X[NDIM]) { TreeC_grow_centered(*nx,X); }
/******************************************************************************/
void treef_reuse_ ()                         { TreeC_reuse(); }
void treef_reuse__()                         { TreeC_reuse(); }
/******************************************************************************/
void treef_approx_grav_ ()                   { TreeC_approx_grav(); }
void treef_approx_grav__()                   { TreeC_approx_grav(); }
/******************************************************************************/
void treef_estimate_rho_ (int*nx)            { TreeC_estimate_rho(*nx); }
void treef_estimate_rho__(int*nx)            { TreeC_estimate_rho(*nx); }
/******************************************************************************/
void treef_estimate_n_ (int*nx)              { TreeC_estimate_n(*nx); }
void treef_estimate_n__(int*nx)              { TreeC_estimate_n(*nx); }
/******************************************************************************/
void treef_sticky_ (INPUT_TYPE*t,int*i1,int*i2,int*ni,int*na,
		    VELIS,INPUT_TYPE*s) { TreeF_ialist(i1,i2,*ni,na,s,*t,VELS); }
void treef_sticky__(INPUT_TYPE*t,int*i1,int*i2,int*ni,int*na,
		    VELIS,INPUT_TYPE*s) { TreeF_ialist(i1,i2,*ni,na,s,*t,VELS); }
/******************************************************************************/
void treef_sph_ (int*i1,int*i2,int*ni,int*na,INPUT_TYPE*s)
                                       { TreeF_ialist(i1,i2,*ni,na,s,-1,ZEROS); }
void treef_sph__(int*i1,int*i2,int*ni,int*na,INPUT_TYPE*s)
                                       { TreeF_ialist(i1,i2,*ni,na,s,-1,ZEROS); }
/******************************************************************************/
float treef_root_center_ (int*I)             { return TreeC_root_center(*I); }
float treef_root_center__(int*I)             { return TreeC_root_center(*I); }
/******************************************************************************/
float treef_root_radius_ ()                  { return TreeC_root_radius(); }
float treef_root_radius__()                  { return TreeC_root_radius(); }
/******************************************************************************/
float treef_current_theta_ (INPUT_TYPE*m)    { return TreeC_current_theta(*m); }
float treef_current_theta__(INPUT_TYPE*m)    { return TreeC_current_theta(*m); }
/******************************************************************************/
float treef_current_eps_ ()                  { return TreeC_current_eps(); }
float treef_current_eps__()                  { return TreeC_current_eps(); }
/******************************************************************************/
int treef_current_kernel_ ()                 { return TreeC_current_kernel(); }
int treef_current_kernel__()                 { return TreeC_current_kernel(); }
/******************************************************************************/
int treef_softening_ ()                      { return TreeC_softening(); }
int treef_softening__()                      { return TreeC_softening(); }
/******************************************************************************/
int treef_no_cells_ ()                       { return TreeC_No_cells(); }
int treef_no_cells__()                       { return TreeC_No_cells(); }
/******************************************************************************/
int treef_depth_ ()                          { return TreeC_depth(); }
int treef_depth__()                          { return TreeC_depth(); }
/******************************************************************************/
void treef_stats_ ()                         { TreeC_stats(); }
void treef_stats__()                         { TreeC_stats(); }
/******************************************************************************/
#undef POSS
#undef VELS
#undef ACCS
#undef POSIS
#undef VELIS
#undef ACCIS
#undef ZEROS
/******************************************************************************
 ******************************************************************************/

int    clock_ ()               { return clock(); }
int    clock__()               { return clock(); }
double drand48_()              { return drand48(); }  
double drand48__()             { return drand48(); }
double clocks_per_second_()    { return CLOCKS_PER_SEC; }
double clocks_per_second__()   { return CLOCKS_PER_SEC; }
void   srand48_(int*S)         { srand48(*S); }
void   srand48__(int*S)        { srand48(*S); }

/******************************************************************************
 ******************************************************************************/
